"""
.. versionadded:: 0.1.14

Code related to heatmaps.
"""
from builtins import range
import logging
import numpy
from .utils import get_grid_figure

import scipy.spatial.distance as distance
import scipy.cluster.hierarchy as hclust

LOG = logging.getLogger(__name__)


def baseheatmap(data, ax, norm=None, cmap=None, xticks=None, yticks=None,
                fontsize=18, meshopts=None, annot=False, annotopts=None):
    """

    .. versionchanged:: 0.2.3
        added *annot* and *annot_args* arguments

    A basic heatmap using :func:`matplotlib.pyplot.pcolormesh`. It expect a
    :class:`pandas.DataFrame`.

    .. note::

        Rows a plot bottom to up, while the columns left to right. Change the
        order of the DataFrame if needed.

    Arguments:
        data (pandas.DataFrame): matrix to plot. The DataFrame labels are used
        ax: axes to use
        norm: if needed, :class:`matplotlib.colors.BoundaryNorm` or
            :class:`matplotlib.colors.Normalize` can be used to fine tune the
            colors
        cmap (None, matplotlib.colors.ListedColormap): color map to use
        xticks (None, dict): dictionary with additional options to pass to
            *set_xticklabels*
        yticks (None, dict): dictionary with additional options to pass to
            *set_yticklabels*
        fontsize (int): font size to use for the labels
        meshopts (None, dict): additional options to pass to
            :func:`matplotlib.pyplot.pcolormesh`
        annot (bool): if True the values of the matrix will be added
        annot_args (None, dict): dictionary with the options for the
            annotations. The option *format* is a function that returns the
            formatted number, defaults to a number with no decimal part

    Returns:
        matplotlib.collections.QuadMesh: the return value of
        :func:`matplotlib.pyplot.pcolormesh`
    """
    mesh_args = {
        'edgecolor': 'face'
    }
    xticks_args = {
        'pos': 'bottom',
        'rotation': 'vertical'
    }
    yticks_args = {
        'pos': 'right',
        'rotation': 'horizontal'
    }
    annot_args = {
        'fontsize': fontsize * 0.75,
        'color': 'k',
        'format': lambda x: "%.0f" % x
    }

    if meshopts is not None:
        mesh_args.update(meshopts)

    if xticks is not None:
        xticks_args.update(xticks)

    if yticks is not None:
        yticks_args.update(yticks)

    if annotopts is not None:
        annot_args.update(annotopts)

    mesh = ax.pcolormesh(data.values, cmap=cmap, norm=norm, **mesh_args)

    ax.xaxis.set_ticks_position(xticks_args['pos'])
    ax.set_xticks(numpy.arange(0.5, len(data.columns), 1))
    ax.set_xticklabels(
        data.columns, fontsize=fontsize, rotation=xticks_args['rotation']
    )

    ax.yaxis.set_ticks_position(yticks_args['pos'])
    ax.set_yticks(numpy.arange(0.5, len(data.index), 1))
    ax.set_yticklabels(
        data.index, fontsize=fontsize, rotation=yticks_args['rotation']
    )

    ax.set_ylim(top=len(data))
    ax.set_xlim(right=len(data.columns))

    if annot:
        for y in range(data.shape[0]):
            for x in range(data.shape[1]):
                ax.text(
                    x + 0.5,
                    y + 0.5,
                    annot_args['format'](data.iloc[y, x]),
                    ha='center',
                    va='center',
                    fontsize=annot_args['fontsize'],
                    color=annot_args['color']
                )

    return mesh


def grouped_spine(groups, labels, ax, which='y', spine='right',
                  spine_opts=None, start=0):
    """
    .. versionchanged:: 0.2.0
        added *va*, *ha* keys to *spine_opts*, changed the label positioning

    .. versionchanged:: 0.2.5
        added *start* parameter

    Changes the spine of an heatmap axis given the groups of labels.

    .. note::

        It should work for any plot, but was not tested

    Arguments:
        groups (iterable): a nested list where each is element is a list
            containing the labels that belong to that group.
        labels (iterable): an iterable with the labels of the groups. Needs to
            be in the same order as groups
        ax: axis to use (same as heatmap)
        which (str): to specify the axis, either *x* or *y*
        spine (str): position of the spine. if *which* is **x** accepted values
            are *top* and *bottom*, if which is **y** *left* and *right* are
            accepted
        spine_opts (dict): additional options to pass to the spine class
        start (int): the start coordinate for the grouped spine. Defaults to 0

    """
    spine_args = dict(
        ec='k',
        position=15,
        group_lw=1.5,
        in_length=13,
        out_length=15,
        fontsize=18,
        va='center',
        ha='left',
        rotation='horizontal'
    )

    if spine_opts is not None:
        spine_args.update(spine_opts)

    major_ticks = [0]
    minor_ticks = []

    for group in groups:
        if not group:
            continue

        major_ticks.append(major_ticks[-1] + len(group))

        group_half = len(group) // 2
        addendum = 0.5

        if len(group) % 2 == 0:
            addendum = 0.
        minor_ticks.append(
            group_half + major_ticks[-2] + addendum
        )

    ax.spines[spine].set_visible(True)
    ax.spines[spine].set_position(('outward', spine_args['position']))
    ax.spines[spine].set_edgecolor(spine_args['ec'])
    ax.spines[spine].set_linewidth(spine_args['group_lw'])

    if which == 'y':
        axis = ax.yaxis
    else:
        axis = ax.xaxis

    axis.set_ticks(numpy.array(major_ticks) + start)
    axis.set_ticks(numpy.array(minor_ticks) + start, minor=True)
    axis.set_ticks_position(spine)
    axis.set_ticklabels(labels, minor=True, va=spine_args['va'],
                        ha=spine_args['ha'], rotation=spine_args['rotation'])
    axis.set_ticklabels([], minor=False)
    axis.set_tick_params(
        direction='in',
        length=spine_args['in_length'],
        which='major',
        width=spine_args['group_lw']
    )
    axis.set_tick_params(
        direction='out',
        length=spine_args['out_length'],
        which='minor',
        width=spine_args['group_lw'],
        pad=spine_args['fontsize'],
        labelsize=spine_args['fontsize']
    )


def dendrogram(data, ax, method='complete', orientation='top', use_dist=True,
               dist_func=distance.pdist):
    """
    .. versionchanged:: 0.1.16
        added *use_dist* and *dist_func* parameters

    Plots a dendrogram of the clustered rows of the given matrix; if the
    columns are to be clustered, the transposed matrix needs to be passed.

    Arguments:
        data (pandas.DataFrame): matrix to plot. The DataFrame labels are used
        ax: axes to use
        method (str): clustering method used, internally
            :func:`scipy.cluster.hierarchy.linkage` is used.
        orientation (str): direction for the plot. *top*, *bottom*, *left* and
            *right* are accepted; *top* will draw the leaves at the bottom.
        use_dist (bool): if True, the function *dist_func* will be applied to
            *data* to get a distance matrix
        dist_func (func): distance function to be used

    Returns:
        The dendrogram plotted, as returned by
        :func:`scipy.cluster.hierarchy.dendrogram`

    """
    if use_dist:
        data = dist_func(data)

    pairwise_dists = distance.squareform(data)

    clusters = hclust.linkage(pairwise_dists, method=method)

    dendrogram = hclust.dendrogram(
        clusters,
        ax=ax,
        link_color_func=lambda x: 'black',
        orientation=orientation
    )
    ax.grid(False)
    ax.set_axis_off()
    ax.set_facecolor('white')
    ax.set_xticks([])
    ax.set_yticks([])

    return dendrogram


def heatmap_clustered(data, figsize=(10, 5), cmap=None, norm=None):
    """
    Plots a heatmap clustered on both rows and columns.

    Arguments:
        data (pandas.DataFrame): matrix to plot. The DataFrame labels are used
        figsize (tuple): passed to :func:`mgkit.plots.utils.get_grid_figure`
        cmap (None, matplotlib.colors.ListedColormap): color map to use
        norm: if needed, :class:`matplotlib.colors.BoundaryNorm` or
            :class:`matplotlib.colors.Normalize` can be used to fine tune the
            colors
    """

    fig, gs = get_grid_figure(
        2,
        2,
        figsize=figsize,
        wspace=0,
        hspace=0,
        width_ratios=[0.25, 1],
        height_ratios=[0.25, 1]
    )
    dendr = dendrogram(data, fig.add_subplot(gs[1, 0]), orientation='left')
    dendc = dendrogram(data.T, fig.add_subplot(gs[0, 1]))

    baseheatmap(
        data.ix[dendr['leaves'], dendc['leaves']],
        fig.add_subplot(gs[1, 1]),
        cmap=cmap,
        norm=norm,
        xticks={'pos': 'bottom'}
    )


__all__ = ['baseheatmap', 'grouped_spine', 'dendrogram', 'heatmap_clustered']



"""
.. versionadded:: 0.3.1

Resampling Utilities
====================

*sample* command
----------------

This command samples from a Fasta or FastQ file, based on a probability defined
by the user (0.001 or 1 / 1000 by default, *-r* parameter), for a maximum number
of sequences (100,000 by default, *-x* parameter). By default 1 sample is
extracted, but as many as desired can be taken, by using the *-n* parameter.

The sequence file in input can be either be passed to the standard input or as
last parameter on the command line. By defult a Fasta is expected, unless the
*-q* parameter is passed.

The *-p* parameter specifies the prefix to be used, and if the output files can
be gzipped using the *-z* parameter.

*sample_stream* command
-----------------------

It works in the same way as *sample*, however the file is sampled only once and
the output is the stdout by default. This can be convenient if streams are a
preferred way to sample the file.

*sync* command
--------------

Used to keep in sync forward and reverse read files in paired-end FASTQ.
The scenario is that the *sample* command was used to resample a FASTQ file,
usually the forward, but we need the reverse as well. In this case, the resampled
file, called *master* is passed to the *-m* option and the input file is
the file that is to be synced (reverse). The input file is scanned until the same header is
found in the master file and when that happens, the sequence is written. The
next sequence is then read from the master file and the process is repeated until all
sequence in the master file are found in the input file. This implies having
the 2 files sorted in the same way, which is what the *sample* command does.

.. note::

    the old casava format is not supported by this command at the moment, as
    it's unusual to find it in SRA or other repositories as well.

*rand_seq* command
------------------

Generate random FastA/Q sequences, allowing the specification of GC content and
number of sequences being coding or random. If the output format chosen is
FastQ, qualities are generated using a decreasing model with added noise. A
constant model can be specified instead with a switch. Parameters such GC,
length and the type of model can be infered by passing a FastA/Q file, with
the quality model fit using a LOWESS (using :func:`mgkit.utils.sequence.extrapolate_model`).
The noise in that case is model as the a normal distribution fitted from the
qualities along the sequence deviating from the fitted LOWSS and scaled back by
half to avoid too drastic changes in the qualities. Also the qualities are
clipped at 40 to avoid compatibility problems with FastQ readers. If inferred,
the model can be saved (as a pickle file) and loaded back for analysis

Changes
-------

.. versionchanged:: 0.3.4
    using *click* instead of *argparse. Now *rand_seq* can save and reload models

.. versionchanged:: 0.3.3
    added *sync*, *sample_stream* and *rand_seq* commnads

"""
from __future__ import division
from builtins import range, zip
import logging
import itertools
import uuid
import click
import numpy
import scipy.stats
import pickle
from tqdm import tqdm
import mgkit
from . import utils
from ..utils import sequence
from mgkit.io import fasta, open_file
from mgkit.io.fastq import load_fastq, write_fastq_sequence, choose_header_type

LOG = logging.getLogger(__name__)


@click.group()
@click.version_option()
@utils.cite_option
def main():
    "Main function"
    pass


def infer_parameters(file_handle, fastq_bool, progress):
    LOG.info("Extrapolating model from file %s", file_handle.name)

    if fastq_bool:
        it = load_fastq(file_handle, num_qual=True)
        quals = []
    else:
        it = fasta.load_fasta(file_handle)

    if progress:
        it = tqdm(it)

    gc_content = []

    length = 0

    for record in it:
        length = max(length, len(record[1]))
        gc_content.append(
            sequence.sequence_gc_content(record[1])
        )
        if fastq_bool:
            quals.append(record[2])

    if fastq_bool:
        model = sequence.extrapolate_model(quals)
    else:
        model = None

    gc_content = numpy.mean(gc_content)

    return length, gc_content, model


@main.command('rand_seq', help='Generates random FastA/Q sequences')
@click.option('-v', '--verbose', is_flag=True)
@click.option('-n', '--num-seqs', default=1000, type=click.INT, show_default=True,
              help='Number of sequences to generate')
@click.option('-gc', '--gc-content', default=.5, type=click.FLOAT, show_default=True,
              help='GC content (defaults to .5 out of 1)')
@click.option('-i', '--infer-params', default=None, type=click.File('rb', lazy=False),
              help='Infer parameters GC content and Quality model from file')
@click.option('-r', '--coding-prop', default=0., type=click.FLOAT, show_default=True,
              help='Proportion of coding sequences')
@click.option('-l', '--length', default=150, type=click.INT, show_default=True,
              help='Sequence length')
@click.option('-d', '--const-model', default=False, is_flag=True,
              help='Use a model with constant qualities + noise')
@click.option('-x', '--dist-loc', default=30., type=click.FLOAT, show_default=True,
              help='Use as the starting point quality')
@click.option('-q', '--fastq', default=False, is_flag=True,
              help='The output file is a FastQ file')
@click.option('-m', '--save-model', default=None, type=click.File('wb', lazy=False),
              help='Save inferred qualities model to a pickle file')
@click.option('-a', '--read-model', default=None, type=click.File('rb', lazy=False),
              help='Load qualities model from a pickle file')
@click.option('--progress', default=False, is_flag=True,
              help="Shows Progress Bar")
@click.argument('output_file', type=click.File('wb', lazy=False), default='-')
def rand_sequence_command(verbose, num_seqs, gc_content, infer_params,
                          coding_prop, length, const_model, dist_loc, fastq,
                          save_model, read_model, progress, output_file):

    mgkit.logger.config_log(level=logging.DEBUG if verbose else logging.INFO)

    if fastq:
        # default values, unless infer_parameters is used
        min_qual = 0
        max_qual = 60
        if const_model:
            LOG.info("Using constant model with loc=%.1f", dist_loc)
            model = sequence.qualities_model_constant(
                length=length,
                loc=dist_loc
            )
        elif infer_params:
            length, gc_content, model = infer_parameters(
                infer_params,
                fastq,
                progress
            )
            min_qual, max_qual = model[2:]
            model = model[:2]
        elif read_model:
            LOG.info('Reading saved model')
            read_model = pickle.load(read_model)
            gc_content = read_model['gc_content']
            lw = read_model['lw']
            length = len(lw)
            model = (
                lw,
                getattr(scipy.stats, read_model['dist_family'])(*read_model['dist'])
            )
            # tries to read the min/max quality params, otherwise keep defaults
            try:
                min_qual = read_model['min_qual']
                max_qual = read_model['max_qual']
            except KeyError:
                pass
        else:
            LOG.info("Using decrease model with loc=%.1f", dist_loc)
            model = sequence.qualities_model_decrease(
                length=length,
                loc=dist_loc
            )

        if save_model is not None:
            LOG.info('Saving model to file (%s)',
                     getattr(save_model, 'name', repr(save_model)))
            pickle.dump(
                dict(lw=model[0], dist=model[1].args, dist_family='norm',
                     gc_content=gc_content, max_qual=max_qual,
                     min_qual=min_qual),
                save_model
            )

    # A C T G
    prob = [
        (1 - gc_content) / 2.,
        gc_content / 2.
    ] * 2

    LOG.info(
        '%d Sequences, with a length of %d - coding proportion: %.1f',
        num_seqs,
        length,
        coding_prop
    )
    LOG.info("Probability A %.2f, C %.2f, T %.2f, G %.2f", *prob)

    num_coding = numpy.round(num_seqs * coding_prop).astype(int)
    seq_it = itertools.chain(
        sequence.random_sequences_codon(n=num_coding, length=length),
        sequence.random_sequences(
            n=num_seqs - num_coding,
            length=length,
            p=prob
        )
    )
    if fastq:
        qual_it = sequence.random_qualities(
            n=num_seqs,
            length=length,
            model=model,
            max_qual=max_qual,
            min_qual=min_qual,
        )
    else:
        qual_it = itertools.repeat(num_seqs)

    if progress:
        qual_it = tqdm(qual_it, total=num_seqs)

    for seq, qual in zip(seq_it, qual_it):
        seq_id = str(uuid.uuid4())
        if fastq:
            write_fastq_sequence(output_file, seq_id, seq, qual)
        else:
            fasta.write_fasta_sequence(output_file, seq_id, seq)


def compare_header(header1, header2, header_type=None):

    if header_type is None:
        return header1[-1] == header2[-1]
    else:
        return header1.split(' ')[0] == header2.split(' ')[0]


@main.command('sync', help="""Syncs a FastQ file generated with *sample* with
              the original pair of files.""")
@click.option('-v', '--verbose', is_flag=True)
@click.option('-m', '--master-file', type=click.File('rb', lazy=False), required=True,
              help='Resampled FastQ file that is out of sync with the original pair')
@click.argument('input-file', type=click.File('rb', lazy=False), default='-')
@click.argument('output-file', type=click.File('wb', lazy=False), default='-')
def fq_sync_command(verbose, master_file, input_file, output_file):

    mgkit.logger.config_log(level=logging.DEBUG if verbose else logging.INFO)

    master_file = load_fastq(master_file, num_qual=False)
    master_header = next(master_file)[0]

    header_type = choose_header_type(master_header)

    written_count = 0

    for header, seq, qual in load_fastq(input_file, num_qual=False):

        if compare_header(master_header, header, header_type):
            write_fastq_sequence(output_file, header, seq, qual)
            written_count += 1
            try:
                master_header = next(master_file)[0]
            except StopIteration:
                break

    LOG.info("Wrote %d FASTQ sequences", written_count)


@main.command('sample', help='Sample a FastA/Q multiple times')
@click.option('-v', '--verbose', is_flag=True)
@click.option('-p', '--prefix', default='sample', show_default=True,
              help='Prefix for the file name(s) in output')
@click.option('-n', '--number', type=click.INT, default=1, show_default=True,
              help='Number of samples to take')
@click.option('-r', '--prob', type=click.FLOAT, default=10**-3,
              show_default=True, help='Probability of picking a sequence')
@click.option('-x', '--max-seq', type=click.INT, default=10**5,
              show_default=True, help='Maximum number of sequences')
@click.option('-q', '--fastq', default=False, is_flag=True,
              help='The input file is a fastq file')
@click.option('-z', '--gzip', is_flag=True, default=False,
              help='gzip output files')
@click.argument('input-file', type=click.File('rb', lazy=False), default='-')
def sample_command(verbose, prefix, number, prob, max_seq, fastq, gzip,
                   input_file):
    mgkit.logger.config_log(level=logging.DEBUG if verbose else logging.INFO)
    LOG.info(
        "Sampling %s file (%d) chunks with prefix (%s)",
        'FastQ' if fastq else 'Fasta',
        number,
        prefix
    )

    if (prob > 1) or (prob <= 0):
        utils.exit_script(
            "The probability value ({}) is outside the correct range" +
            " (0 < p <= 1)",
            1
        )

    dist = scipy.stats.binom(1, prob)

    LOG.info(
        "Probability of picking a sequence (%.5f), max number of seqs %d",
        prob,
        max_seq
    )
    name_mask = "%s-{0:05}.%s" % (prefix, 'fq' if fastq else 'fa')

    if gzip:
        name_mask += '.gz'
        LOG.info("Output files will be compressed (gzip)")

    output_files = [
        dict(
            h=open_file(name_mask.format(i), 'wb'),
            c=0
        )
        for i in range(number)
    ]

    load_func = load_fastq if fastq else fasta.load_fasta
    write_func = write_fastq_sequence if fastq else fasta.write_fasta_sequence

    for seq in load_func(input_file):
        # reached the maximum number of sequences for all samples
        if all(x['c'] == max_seq for x in output_files):
            break

        for output in output_files:
            if output['c'] == max_seq:
                continue

            if dist.rvs():
                write_func(output['h'], *seq)
                output['c'] += 1


@main.command('sample_stream', help="""Samples a FastA/Q one time, alternative
              to sample if multiple sampling is not needed""")
@click.option('-v', '--verbose', is_flag=True)
@click.option('-r', '--prob', type=click.FLOAT, default=10**-3,
              help='Probability of picking a sequence')
@click.option('-x', '--max-seq', type=click.INT, default=10**5,
              help='Maximum number of sequences')
@click.option('-q', '--fastq', default=False, is_flag=True,
              help='The input file is a fastq file')
@click.argument('input-file', type=click.File('rb', lazy=False), default='-')
@click.argument('output-file', type=click.File('wb', lazy=False), default='-')
def stream_sample_command(verbose, prob, max_seq, fastq, input_file,
                          output_file):
    mgkit.logger.config_log(level=logging.DEBUG if verbose else logging.INFO)

    if (prob > 1) or (prob <= 0):
        utils.exit_script(
            "The probability value ({}) is outside the correct range" +
            " (0 < p <= 1)",
            1
        )

    dist = scipy.stats.binom(1, prob)

    LOG.info(
        "Probability of picking a sequence (%.5f), max number of seqs %d",
        prob,
        max_seq
    )

    load_func = load_fastq if fastq else fasta.load_fasta
    write_func = write_fastq_sequence if fastq else fasta.write_fasta_sequence

    count = 0

    for index, seq in enumerate(load_func(input_file)):
        # reached the maximum number of sequences for all samples
        if count >= max_seq:
            LOG.info('Read the first %d sequences', index + 1)
            break

        if dist.rvs():
            write_func(output_file, *seq)
            count += 1



"""
Calculates pN/pS values
***********************

This script calculates pN/pS using the data produced by the script
:ref:`snp_parser`. The result table is a CSV file.

Calculate Rank pN/pS
********************

The `rank` command of the script reads SNPs information and calculate for each
element of a specific taxonomic rank (species, genus, family, etc.) its pN/pS.
Another option is the *None* rank, which makes the script use the taxonomic ID
found in the annotations.

For example, choosing the rank *genus* a table will be produced, similar to::

    Prevotella,0.0001,1,1.1,0.4
    Methanobrevibacter,1,0.5,0.6,0.8

A pN/pS value for each genus and sample (4 in this case) will be calculated.

It is important to specify the taxonomic IDs to include in tha calculations. By
default only bacteria are included. To get those values, the taxonomy can be
queried using `taxon-utils get`.

Calculate Gene/Rank pN/pS
*************************

The `full` command create a gene/taxon table of pN/pS, internally is a pandas
MultiIndex DataFrame, written in CSV format after script execution. The
difference with the `rank` is the pN/pS calculation is for each gene/taxon and
by default the `gene_id` from the original GFF file is used (which is stored
in the file generated by `snp_parser`). If other gene IDs needs to be used, a
table file can be provided, which can be passed in two column formats.

The default in MGKit is to use Uniprot gene IDs for the functions, but we may
want to examine the Kegg Orthologs instead. A table can be passed where the
first column in the gene_id stored in the GFF file and the second is the KO::

    Q7N6F9  K05685
    Q7N6F9  K01242
    G7E4F2  K05625

The `Q7N6F9` gene_id is repeated because it has multiple correspondences to KOs
and this format needs to be selected using the `-2` option of the command.

The default type of table expected by the command is a table with a gene ID as
first column one or more tab separated columns with mappings. The previous
table would look like this::

    Q7N6F9  K05685  K01242
    G7E4F2  K05625

These tables can be created from the original GFF file, assuming that mappings
to KO, EC Numbers are included, with a command line like this::

     edit-gff view -a gene_id -a map_KO final.contigs-a3.gff.gz | tr ',' '\t'

Extracting the KOs (which are comma separated in a MGKit GFF file) and changing
any comma to *tab*. This table can be passed to the script and will make it
possible to calculate the pN/pS for the KOs associated to the genes. Only gene
IDs present in this file have a calculated pN/pS.

Normally you combine the all isoforms with the same the *gene_id* to produce a
single pN/pS, but if it's needed, the `-u` option can be used to calculate a
pN/pS for each line in the GFF file.

Changes
*******

.. versionchanged:: 0.5.1
    bug fix

.. versionchanged:: 0.5.1

    * added option to include the lineage as a string
    * added option to use the *uids* from the GFF instead of *gene_id*, this
      does not require the GFF file, they are embedded into the *.pickle* file

.. versionadded:: 0.5.0
"""


import logging
import click
import pickle
import functools
from mgkit import taxon
from mgkit.snps import conv_func
import mgkit.snps.mapper
import mgkit.snps.funcs
import mgkit.snps.filter as snp_filter
from . import utils
from .. import logger

LOG = logging.getLogger(__name__)

# changing the value for the script to remove subclass as rank
taxon.TAXON_RANKS = tuple(rank for rank in taxon.TAXON_RANKS
                          if not rank.startswith('sub'))


def get_lineage(taxonomy, taxon_id):
    if taxon_id not in taxonomy:
        return taxon_id
    return taxonomy.get_lineage_string(taxon_id, only_ranked=True,
                                       with_last=True,
                                       sep=';')


@click.group()
@click.version_option()
@utils.cite_option
def main():
    "Main function"
    pass


@main.command('rank', help="""Calculates pN/pS for a taxonomic rank""")
@click.option('-v', '--verbose', is_flag=True)
@click.option('-t', '--taxonomy', type=click.File('rb', lazy=False),
              help="Taxonomy file", required=True)
@click.option('-s', '--snp-data', type=click.File('rb', lazy=False),
              help="SNP data, output of `snp_parser`", required=True)
@click.option('-r', '--rank', default='order', help='Taxonomic rank',
              type=click.Choice(taxon.TAXON_RANKS + ('None',),
                                case_sensitive=False), show_default=True)
@click.option('-m', '--min-num', default=2, type=click.IntRange(min=2),
              help='Minimum number of samples with a pN/pS to accept',
              show_default=True)
@click.option('-c', '--min-cov', default=4, type=click.IntRange(min=1),
              help='Minimum coverage for SNPs to be accepted',
              show_default=True)
@click.option('-i', '--taxon_ids', type=click.INT, multiple=True,
              help='Taxon IDs to include', default=(2,), show_default=True)
@click.option('-u', '--unstack', is_flag=True, show_default=True,
              help='Samples are not in columns but as an array')
@click.option('-l', '--lineage', is_flag=True, show_default=True,
              help='Use lineage string instead of taxon_id')
@click.option('-ps', '--only-ps', is_flag=True, show_default=True,
              help='Only calculate pS values')
@click.option('-pn', '--only-pn', is_flag=True, show_default=True,
              help='Only calculate pN values')
@click.argument('txt_file', type=click.File('w', lazy=False), default='-')
def gen_rank(verbose, taxonomy, snp_data, rank, min_num, min_cov,
             taxon_ids, unstack, lineage, only_ps, only_pn, txt_file):

    logger.config_log(level=logging.DEBUG if verbose else logging.INFO)

    taxonomy = taxon.Taxonomy(taxonomy)

    LOG.info('Only taxa below %s will be included', ', '.join(taxonomy[taxon_id].s_name for taxon_id in taxon_ids))
    LOG.info('Rank "%s" and above will be included', rank)

    snp_data = pickle.load(snp_data)

    filters = snp_filter.get_default_filters(taxonomy, min_cov=min_cov,
                                             include_only=taxon_ids)

    if rank not in taxon.TAXON_RANKS:
        rank = None
    
    if rank is None:
        taxon_func = None
    else:
        taxon_func = functools.partial(
            mgkit.snps.mapper.map_taxon_id_to_rank,
            taxonomy=taxonomy,
            rank=rank
        )
    
    partial_calc = False
    partial_syn = True
    if only_ps or only_pn:
        partial_calc = True
        if only_pn:
            partial_syn = False
        LOG.info('Only "%s" will be calculated ', "pS"  if partial_syn else "pN")

    pnps = mgkit.snps.funcs.combine_sample_snps(snp_data, min_num, filters, index_type='taxon',
                                                taxon_func=taxon_func, partial_calc=partial_calc,
                                                partial_syn=partial_syn)

    if lineage:
        pnps.rename(lambda x: get_lineage(taxonomy, x), inplace=True)

    pnps.index.names = ['taxon']

    if unstack:
        pnps = pnps.unstack()

    pnps.to_csv(txt_file)


def read_gene_map_default(file_handle, separator):

    gene_map = {}

    for line in file_handle:
        fields = line.rstrip().split(separator)
        gene_map[fields[0]] = set(fields[1:])

    return gene_map


def read_gene_map_two_columns(file_handle, separator):

    gene_map = {}

    for line in file_handle:
        key, value = line.rstrip().split(separator)
        try:
            gene_map[key].add(value)
        except KeyError:
            gene_map[key] = set([value])

    return gene_map


@main.command('full', help="""Calculates pN/pS""")
@click.option('-v', '--verbose', is_flag=True)
@click.option('-t', '--taxonomy', type=click.File('rb', lazy=False),
              help="Taxonomy file", required=True)
@click.option('-s', '--snp-data', type=click.File('rb', lazy=False),
              help="SNP data, output of `snp_parser`", required=True)
@click.option('-r', '--rank', default='order', help='Taxonomic rank',
              type=click.Choice(taxon.TAXON_RANKS + ('None',),
                                case_sensitive=False), show_default=True)
@click.option('-m', '--min-num', default=2, type=click.IntRange(min=2),
              help='Minimum number of samples with a pN/pS to accept',
              show_default=True)
@click.option('-c', '--min-cov', default=4, type=click.IntRange(min=1),
              help='Minimum coverage for SNPs to be accepted',
              show_default=True)
@click.option('-i', '--taxon_ids', type=click.INT, multiple=True,
              help='Taxon IDs to include', default=(2,), show_default=True)
@click.option('-u', '--use-uid', is_flag=True, show_default=True,
              help='Use uids from the GFF file instead of gene_id as genes')
@click.option('-g', '--gene-map', type=click.File(mode='r', lazy=False),
              help='Dictionary to map *gene_id* to another ID', default=None)
@click.option('-2', '--two-columns', is_flag=True,
              help='gene-map is a two columns table with repeated keys')
@click.option('-p', '--separator', default='\t', show_default=True,
              help='column separator for gene-map file')
@click.option('-l', '--lineage', is_flag=True, show_default=True,
              help='Use lineage string instead of taxon_id')
@click.option('-ps', '--only-ps', is_flag=True, show_default=True,
              help='Only calculate pS values')
@click.option('-pn', '--only-pn', is_flag=True, show_default=True,
              help='Only calculate pN values')
@click.argument('txt_file', type=click.File('w', lazy=False), default='-')
def gen_full(verbose, taxonomy, snp_data, rank, min_num, min_cov,
             taxon_ids, use_uid, gene_map, two_columns, separator, lineage,
             only_ps, only_pn, txt_file):

    logger.config_log(level=logging.DEBUG if verbose else logging.INFO)

    if gene_map is not None:
        LOG.info('Reading gene-map')
        if two_columns:
            gene_map = read_gene_map_two_columns(gene_map, separator)
        else:
            gene_map = read_gene_map_default(gene_map, separator)
        
        gene_map = functools.partial(
            mgkit.snps.mapper.map_gene_id,
            gene_map=gene_map
        )

    taxonomy = taxon.Taxonomy(taxonomy)

    LOG.info('Only taxa below %s will be included', ', '.join(taxonomy[taxon_id].s_name for taxon_id in taxon_ids))
    LOG.info('Rank "%s" and above will be included', rank)

    snp_data = pickle.load(snp_data)

    if rank not in taxon.TAXON_RANKS:
        rank = None
    
    if rank is None:
        taxon_func = None
    else:
        taxon_func = functools.partial(
            mgkit.snps.mapper.map_taxon_id_to_rank,
            taxonomy=taxonomy,
            rank=rank
        )

    filters = snp_filter.get_default_filters(taxonomy, min_cov=min_cov,
                                             include_only=taxon_ids)

    partial_calc = False
    partial_syn = True
    if only_ps or only_pn:
        partial_calc = True
        if only_pn:
            partial_syn = False
        LOG.info('Only "%s" will be calculated ', "pS"  if partial_syn else "pN")

    pnps = mgkit.snps.funcs.combine_sample_snps(snp_data, min_num, filters, gene_func=gene_map,
                                                taxon_func=taxon_func, use_uid=use_uid,
                                                partial_calc=partial_calc, partial_syn=partial_syn)

    if lineage:
        pnps.rename(lambda x:  get_lineage(taxonomy, x), inplace=True)

    pnps.index.names = ['gene', 'taxon']

    pnps.to_csv(txt_file)



import pytest
import requests
from ftplib import FTP
import tarfile
import io
import os
from mgkit.taxon import Taxonomy

try:
    requests.get('http://www.google.com')
    conn_ko = False
except requests.exceptions.ConnectionError:
    conn_ko = True

# To disable tests that require a remote connection
if os.environ.get('MGKIT_TESTS_CONN_SKIP', False):
    conn_ko = True

skip_no_connection = pytest.mark.skipif(conn_ko, reason='No connection available')


@pytest.fixture(scope='session')
def taxonomy_files(tmpdir_factory):
    if conn_ko:
        return None

    taxdump_dir = tmpdir_factory.mktemp('taxdump')
    tax_file = (taxdump_dir / 'taxdump.tar.gz').strpath
    # Seems like the NCBI FTP is not FTP anymore, pointing to the new resource
    url = 'https://ftp.ncbi.nlm.nih.gov/pub/taxonomy/taxdump.tar.gz'
    file_content = requests.get(url)
    open(tax_file, 'wb').write(file_content.content)
    # old code (for FTP)
    # ftp = FTP('ftp.ncbi.nlm.nih.gov')
    # ftp.login()
    # ftp.cwd('/pub/taxonomy/')
    # tax_handle = io.open(tax_file, 'wb')
    # ftp.retrbinary('RETR taxdump.tar.gz', tax_handle.write, 1000)
    # ftp.quit()
    # tax_handle.close()

    tarfile.open(tax_file, mode='r|gz').extractall(
        taxdump_dir.strpath
    )

    return [
        (taxdump_dir / file_name).strpath
        for file_name in ('nodes.dmp', 'names.dmp', 'merged.dmp')
    ]


@skip_no_connection
@pytest.fixture(scope='session')
def ncbi_taxonomy(taxonomy_files):
    taxonomy = Taxonomy()
    taxonomy.read_from_ncbi_dump(*taxonomy_files)
    return taxonomy


EGGNOG_V3_FILES = {
    'members': 'http://eggnog.embl.de/version_3.0/data/downloads/NOG.members.txt.gz',
    'description': 'http://eggnog.embl.de/version_3.0/data/downloads/NOG.description.txt.gz',
    'funccat': 'http://eggnog.embl.de/version_3.0/data/downloads/NOG.funccat.txt.gz',
}


@skip_no_connection
@pytest.fixture(scope='session')
def eggnog_v3(tmpdir_factory):
    eggnog_v3_files_dir = tmpdir_factory.mktemp('eggnog_v3')
    file_paths = {}
    for key, url in EGGNOG_V3_FILES.items():
        file_content = requests.get(url)
        file_path = (eggnog_v3_files_dir / "{}.gz".format(key)).strpath
        file_paths[key] = file_path
        open(file_path, 'wb').write(file_content.content)
    return file_paths


# Same for EXPASY, not anymore an FTP file, like for NCBI
# EXPASY_SERVER = "ftp.expasy.org"
# EXPASY_DIR = "/databases/enzyme/"
EXPASY_FILE = "enzclass.txt"
EXPASY_FILE_URL = 'https://ftp.expasy.org/databases/enzyme/enzclass.txt'
EXPASY_DAT = "enzyme.dat"
EXPASY_DAT_URL = 'https://ftp.expasy.org/databases/enzyme/enzyme.dat'


@skip_no_connection
@pytest.fixture(scope='session')
def expasy_files(tmpdir_factory):
    expasy_file_dir = tmpdir_factory.mktemp('expasy')
    expasy_file_path = (expasy_file_dir / EXPASY_FILE).strpath
    expasy_dat_path = (expasy_file_dir / EXPASY_DAT).strpath

    file_content = requests.get(EXPASY_FILE_URL)
    open(expasy_file_path, 'wb').write(file_content.content)

    file_content = requests.get(EXPASY_DAT_URL)
    open(expasy_dat_path, 'wb').write(file_content.content)

    # old code for reference
    # ftp = FTP(EXPASY_SERVER)
    # ftp.login()
    # ftp.cwd(EXPASY_DIR)
    # expasy_file_handle = io.open(expasy_file_path, 'wb')
    # ftp.retrbinary('RETR {}'.format(EXPASY_FILE), expasy_file_handle.write, 1000)
    # expasy_file_handle.close()
    # expasy_dat_handle = io.open(expasy_dat_path, 'wb')
    # ftp.retrbinary('RETR {}'.format(EXPASY_DAT), expasy_dat_handle.write, 1000)
    # expasy_dat_handle.close()
    # ftp.quit()

    return expasy_file_path, expasy_dat_path



import pytest
from conftest import skip_no_connection
from mgkit.net.pfam import get_pfam_families

@skip_no_connection
def test_pfam_families_id():
    assert 'Protamine_P1' in get_pfam_families(key='id')


@skip_no_connection
def test_pfam_families_ac():
    assert 'PF07938' in get_pfam_families(key='ac')



"""
Module used to map taxon_id to different levels in the taxonomy.
"""


def map_taxon_by_id_list(taxon_id, map_ids, func):
    """
    Maps a taxon_id to a list of taxon IDs, using the function supplied.

    Arguments:
        taxon_id (int): taxon ID to map
        map_ids (iterable): list of taxon IDs to which the taxon_id will be
            mapped.
        func (func): function used to map the IDs, accepts two taxon IDs

    Results:
        generator: generator expression of all IDs in map_ids to which taxon_id
            can be mapped.

    Example:
        If mapping a taxon (Prevotella ruminicola) to Prevotella or
        Clostridium, using as `func` :func:`mgkit.taxon.is_ancestor` and
        taxonomy is an instance of :class:`mgkit.taxon.Taxonomy`.

        >>> import functools
        >>> from mgkit.taxon import is_ancestor
        >>> func = functools.partial(is_ancestor, taxonomy)
        >>> list(map_taxon_by_id_list(839, [838, 1485], func))
        [838]

    """
    taxon_ids = (
        map_id
        for map_id in map_ids
        if func(taxon_id, map_id)
    )

    return taxon_ids



import pytest

from mgkit.io import fasta
from mgkit.io.blast import parse_blast_tab, parse_uniprot_blast, \
    parse_accession_taxa_table

@pytest.fixture
def blast_tab_file(shared_datadir):
    return str(shared_datadir / 'blast-outfmt6.tsv')


@pytest.fixture
def blast_seq(shared_datadir):
    return dict(
        fasta.load_fasta_rename(str(shared_datadir / 'blast-seq.fa'))
    )


@pytest.fixture
def ncbi_taxa_file(shared_datadir):
    return str(shared_datadir / 'ncbi-taxa.tab')


@pytest.fixture
def blast_lengths(blast_seq):
    return {
        seq_id: len(blast_seq[seq_id])
        for seq_id in blast_seq
    }

def test_parse_blast_tab1(blast_tab_file):
    seq_id, hit = next(parse_blast_tab(blast_tab_file))
    assert seq_id == 'M95099.1'


def test_parse_blast_tab2(blast_tab_file):
    seq_id, hit = next(parse_blast_tab(blast_tab_file))
    assert len(hit) == 6


def test_parse_blast_tab3(blast_tab_file):
    seq_id, hit = next(parse_blast_tab(blast_tab_file))
    assert hit[-1] ==  '14520'


def test_parse_blast_tab4(blast_tab_file):
    seq_id, hit = next(parse_blast_tab(blast_tab_file, key_func=lambda x: x.split('.')[0]))
    assert seq_id == 'M95099'


def test_parse_blast_tab5(blast_tab_file):
    seq_id, hit = next(parse_blast_tab(
        blast_tab_file,
        key_func=lambda x: x.split('.')[0],
        ret_col=None)
    )
    assert len(hit) == 12


def test_parse_uniprot_blast1(blast_tab_file):
    key_func = lambda x: x.split('.')[0]

    a = next(parse_uniprot_blast(blast_tab_file, name_func=key_func))

    assert a.gene_id == 'M95099'


def test_parse_uniprot_blast2(blast_tab_file):
    key_func = lambda x: x.split('.')[0]

    a = next(parse_uniprot_blast(blast_tab_file, name_func=key_func))

    assert a.bitscore == 14520


def test_parse_uniprot_blast3(blast_tab_file):
    key_func = lambda x: x.split('.')[0]

    bts = [
        a.bitscore
        for a in parse_uniprot_blast(blast_tab_file, name_func=key_func, bitscore=170)
    ]

    assert bts == [14520]


def test_parse_uniprot_blast4(blast_tab_file, blast_lengths):
    key_func = lambda x: x.split('.')[0]

    a = next(parse_uniprot_blast(blast_tab_file, name_func=key_func, seq_lengths=blast_lengths))

    assert a.phase == 0


def test_parse_accession_taxa_table1(ncbi_taxa_file):
    assert next(parse_accession_taxa_table(ncbi_taxa_file, key=0, value=1,
                                           no_zero=False))[0] == 'KH113978'


def test_parse_accession_taxa_table2(ncbi_taxa_file):
    assert next(parse_accession_taxa_table(ncbi_taxa_file, key=0, value=1,
                                           no_zero=False))[1] == 0


def test_parse_accession_taxa_table3(ncbi_taxa_file):
    assert next(parse_accession_taxa_table(ncbi_taxa_file, key=0, value=1,
                                           no_zero=True))[0] == 'XM_006695435'


def test_parse_accession_taxa_table4(ncbi_taxa_file):
    acc_ids = ['XM_006695435']
    iterator = parse_accession_taxa_table(ncbi_taxa_file, key=0, value=1,
                                          acc_ids=acc_ids)
    iterator = list(acc_id for acc_id, taxon_id in iterator)
    assert acc_ids == iterator


def test_parse_accession_taxa_table5(ncbi_taxa_file):
    acc_ids = {'XM_006695435'}
    iterator = parse_accession_taxa_table(ncbi_taxa_file, key=0, value=1,
                                          acc_ids=acc_ids)
    iterator = list(acc_id for acc_id, taxon_id in iterator)
    assert list(acc_ids) == iterator



import pytest
from conftest import skip_no_connection, ncbi_taxonomy

import functools
from mgkit.mappings.taxon import map_taxon_by_id_list
from mgkit.mappings.enzyme import parse_expasy_file, parse_expasy_dat
from mgkit.taxon import is_ancestor


@skip_no_connection
def test_map_taxon_by_id_list1(ncbi_taxonomy):
    result = map_taxon_by_id_list(
        839,
        [838, 1485],
        functools.partial(
            is_ancestor,
            ncbi_taxonomy
        )
    )
    assert list(result) == [838]


@skip_no_connection
def test_map_taxon_by_id_list2(ncbi_taxonomy):
    result = map_taxon_by_id_list(
        2172,
        [838, 1485],
        functools.partial(
            is_ancestor,
            ncbi_taxonomy
        )
    )
    assert list(result) == []


@skip_no_connection
def test_expasy_file_read(expasy_files):
    expasy_file, _ = expasy_files
    assert parse_expasy_file(expasy_file)['1.1'] == 'Acting on the CH-OH group of donors'


@skip_no_connection
def test_expasy_dat_read1(expasy_files):
    _ , expasy_dat = expasy_files
    entries = parse_expasy_dat(expasy_dat)
    for entry in entries:
        if entry['ID'][0] == '1.1.1.121':
            assert entry['CA'][0] == 'D-aldose + NAD(+) = D-aldonolactone + NADH'


@skip_no_connection
def test_expasy_dat_read2(expasy_files):
    _ , expasy_dat = expasy_files
    entries = parse_expasy_dat(expasy_dat)
    for entry in entries:
        if entry['ID'][0] == '1.1.1.96':
            assert entry['CA'][0] == '3-(3,5-diiodo-4-hydroxyphenyl)lactate + NAD(+) = 3-(3,5-diiodo-4-hydroxyphenyl)pyruvate + NADH'



"""
Fastq utility functions
"""
import re
import numpy
import logging

from .utils import open_file

LOG = logging.getLogger(__name__)

CASAVA_HEADER_OLD = r"""(?P<machine>\w+-\w+):
        (?P<lane>\d):
        (?P<tile>\d+):
        (?P<xcoord>\d+):
        (?P<ycoord>\d+)\#
        (?P<index>(\d|[ACTGN]{6}))/
        (?P<mate>(1|2))"""
"Old casava header regex"

CASAVA_HEADER_NEW = r"""(?P<machine>[\w-]+):
        (?P<runid>\d+):
        (?P<cellid>\w+):
        (?P<lane>\d):
        (?P<tile>\d+):
        (?P<xcoord>\d+):
        (?P<ycoord>\d+)
        [_ ](?P<mate>\d): # underscore for data from from www.ebi.ac.uk/ena/
        (?P<filter>[YN]):
        (?P<bits>\d+):
        (?P<index>[ACTGN+]+)"""
"New casava header regex, including indices for both forward and reverse"

CASAVA_KHMER = r"""
    (?P<machine>[\w-]+):
    (?P<runid>\d+):
    (?P<cellid>\w+):
    (?P<lane>\d):
    (?P<tile>\d+):
    (?P<xcoord>\d+):
    (?P<ycoord>\d+)
    /(?P<mate>\d)
"""


def check_fastq_type(qualities):
    """
    Trys to guess the type of quality string used in a Fastq file

    :param str qualities: string with the quality scores as in the Fastq file
    :return str: a string with the guessed quality score

    .. note::

        Possible values are the following, classified but the values usually
        used in other softwares:

        * ASCII33: sanger, illumina-1.8
        * ASCII64: illumina-1.3, illumina-1.5, solexa-old

    """
    qualities = [ord(char) for char in qualities]

    max_qual = max(qualities)
    min_qual = min(qualities)

    if (min_qual >= 33) and (max_qual <= 73):
        return 'sanger'
    elif (min_qual >= 33) and (max_qual <= 74):
        return 'illumina-1.8'
    elif (min_qual >= 64) and (max_qual <= 104):
        return 'illumina-1.5'
    elif (min_qual >= 67) and (max_qual <= 104):
        return 'illumina-1.3'
    elif (min_qual >= 59) and (max_qual <= 104):
        return 'solexa-old'


def convert_seqid_to_new(seq_id):
    """
    Convert old seq_id format for Illumina reads to the new found in Casava
    1.8+

    :param str seq_id: seq_id of the sequence (stripped of '@')
    :return str: the new format seq_id

    .. note::

        Example from Wikipedia::

            old casava seq_id:
            @HWUSI-EAS100R:6:73:941:1973#0/1
            new casava seq_id:
            @EAS139:136:FC706VJ:2:2104:15343:197393 1:Y:18:ATCAC

    """
    machine_id, cell_lane, tile_num, x_coord, other = seq_id.split(':')
    y_coord, other = other.split('#')
    idx_multiplex, mate = other.split('/')

    return "{0}:0:FC706VJ:{1}:{2}:{3}:{4} {5}:N:0:{6}".format(
        machine_id,
        cell_lane,
        tile_num,
        x_coord,
        y_coord,
        mate,
        idx_multiplex
    )


def convert_seqid_to_old(seq_id, index_as_seq=True):
    """

    .. deprecated:: 0.3.3

    Convert old seq_id format for Illumina reads to the new found in Casava
    until 1.8, which marks the new format.

    :param str seq_id: seq_id of the sequence (stripped of '@')
    :param bool index_as_seq: if True, the index for the multiplex we'll be
        the sequence found at the end of the new format seq_id. Otherwise, 0
        we'll be used
    :return str: the new format seq_id
    """
    left_part, right_part = seq_id.split(' ')
    machine_id, run_id, cell_id, cell_lane, tile_num, x_coord, \
        y_coord = left_part.split(':')

    mate, fail_filter, control_bits, index_seq = right_part.split(':')

    return "HWUSI-{0}:{1}:{2}:{3}:{4}#{5}/{6}".format(
        machine_id,
        cell_lane,
        tile_num,
        x_coord,
        y_coord,
        index_seq if index_as_seq else '0',
        mate
    )


def write_fastq_sequence(file_handle, name, seq, qual, write_mode='a'):
    """

    .. versionchanged:: 0.3.3
        if *qual* is not a string it's converted to chars (phred33)

    Write a fastq sequence to file. If the *file_handle* is a string, the file
    will be opened using *write_mode*.

    :param file_handle: file handle or string.
    :param str name: header to write for the sequence
    :param str seq: sequence to write
    :param str qual: quality string
    """
    file_handle = open_file(file_handle, write_mode)

    if isinstance(qual[0], (int, numpy.integer)):
        qual = ''.join(chr(q + 33) for q in qual)

    file_handle.write(
        "@{name}\n{seq}\n+\n{qual}\n".format(
            name=name,
            seq=seq,
            qual=qual
        ).encode('ascii')
    )


def choose_header_type(seq_id):
    """
    Return the guessed compiled regular expression
    :param str seq_id: sequence header to test

    :return: compiled regular expression object or None if no match found
    """
    if re.search(CASAVA_HEADER_OLD, seq_id, re.X) is not None:
        return re.compile(CASAVA_HEADER_OLD, re.X)
    if re.search(CASAVA_HEADER_NEW, seq_id, re.X) is not None:
        return re.compile(CASAVA_HEADER_NEW, re.X)
    if re.search(CASAVA_KHMER, seq_id, re.X) is not None:
        return re.compile(CASAVA_KHMER, re.X)
    return None


def load_fastq(file_handle, num_qual=False):
    """
    .. versionadded:: 0.3.1

    Loads a fastq file and returns a generator of tuples in which the first
    element is the name of the sequence, the second the sequence and the third
    the quality scores (converted in a numpy array if *num_qual* is True).

    .. note::

        this is a simple parser that assumes each sequence is on 4 lines,
        1st and 3rd for the headers, 2nd for the sequence and 4th the quality
        scores

    Arguments:
        file_handle (str, file): fastq file to open, can be a file name or a
            file handle
        num_qual (bool): if *False* (default), the quality score will be
            returned as ASCII character, if *True* a numpy array.

    Yields:
        tuple: first element is the sequence name/header, the second element is
        the sequence, the third is the quality score. The quality scores are
        kept as a string if *num_qual* is False (default) and converted to a
        numpy array with correct values (0-41) if *num_qual* is True

    Raises:
        ValueError: if the headers in both sequence and quality scores are not
        valid. This implies that the sequence/qualities have carriage returns
        or the file is truncated.

        TypeError: if the qualities are in a format different than sanger
        (min 0, max 40) or illumina-1.8 (0, 41)
    """

    file_handle = open_file(file_handle, 'r')

    if getattr(file_handle, 'name', None) is not None:
        LOG.info("Reading fastq file (%s)", file_handle.name)

    check_qual = True

    sequence_count = 0

    while True:
        header1 = file_handle.readline().decode('ascii').strip()
        # Reached the end of the file
        if not header1:
            break

        seq = file_handle.readline().decode('ascii').strip()

        header2 = file_handle.readline().decode('ascii').strip()
        qualities = file_handle.readline().decode('ascii').strip()

        if (header1[0] != '@') or (header2[0] != '+'):
            raise ValueError(
                "The sequence and quality headers are not valid '{}' != '{}'".format(header1, header2)
            )

        header1 = header1[1:]
        header2 = header2[1:]

        if check_qual:
            qual_type = check_fastq_type(qualities)
            if qual_type not in ('sanger', 'illumina-1.8'):
                raise TypeError(
                    'Quality scores are in "{}" format. You must convert them to Sanger (phred33)'.format(qual_type)
                )
            check_qual = False

        if num_qual:
            qualities = numpy.fromiter((ord(x) for x in qualities), numpy.int) - 33

        sequence_count += 1

        yield header1, seq, qualities

    LOG.info("Read %d fastq sequences", sequence_count)


def load_fastq_rename(file_handle, num_qual=False, name_func=None):
    """
    .. versionadded:: 0.3.3

    Mirrors the same functionality in :func:`mgkit.io.fasta.load_fasta_rename`.
    Renames the header of the sequences using *name_func*, which is called on
    each header. By default, the behaviour is to keep the header to the left of
    the first space (BLAST behaviour).
    """

    if name_func is None:
        def name_func(header): return header.split(' ')[0]

    for header, seq, qual in load_fastq(file_handle, num_qual=num_qual):
        yield name_func(header), seq, qual



"""
This modules define classes and function related to manipulation of GFF/GTF
files.
"""
from __future__ import print_function, division
from builtins import object, zip, bytes, range
import sys
from io import IOBase
from future.utils import viewitems
import random
import itertools
import logging
import uuid
import json
from collections import OrderedDict
try:
    from urllib import unquote, quote
except ImportError:
    # Python3
    from urllib.parse import unquote, quote
import mgkit.io
from ..utils import sequence as seq_utils
from ..consts import MIN_COV
from ..utils.common import between, union_range, ranges_length
from ..utils.trans_tables import UNIVERSAL

LOG = logging.getLogger(__name__)


class AttributeNotFound(Exception):
    """
    Raised if an attribute is not found in a GFF file
    """
    pass


def write_gff(annotations, file_handle, verbose=True):
    """
    .. versionchanged:: 0.1.12
        added *verbose* argument

    Write a GFF to file

    Arguments:
        annotations (iterable): iterable that returns :class:`GFFKegg`
            or :class:`Annotation` instances
        file_handle (str, file): file name or file handle to write to
        verbose (bool): if True, a message is logged
    """

    if isinstance(file_handle, str):
        file_handle = mgkit.io.open_file(file_handle, 'wb')

    if verbose:
        LOG.info(
            "Writing annotations to file (%s)",
            getattr(file_handle, 'name', repr(file_handle))
        )

    for annotation in annotations:
        annotation.to_file(file_handle)


class GenomicRange(object):
    """
    Defines a genomic range

    .. versionchanged:: 0.2.1
        using __slots__ for better memory usage
    """

    __slots__ = ('seq_id', 'strand', 'start', 'end')

    # seq_id = 'None'
    # "Sequence ID"
    # strand = '+'
    # "Strand"
    # start = None
    # "Start of the range, 1-based"
    # end = None
    # "End of the range 1-based"

    def __init__(self, seq_id='None', start=1, end=1, strand='+'):
        self.seq_id = seq_id
        self.strand = strand
        self.start = start
        self.end = end

    def __eq__(self, other):
        if (self.seq_id != other.seq_id) or (self.strand != other.strand):
            return False
        if (self.start != other.start) or (self.end != other.end):
            return False
        return True

    def __len__(self):
        return self.end - self.start + 1

    def __str__(self):
        return "{0}({1}):{2}-{3}".format(
            self.seq_id,
            self.strand,
            self.start,
            self.end
        )

    def __repr__(self):
        return str(self)

    def union(self, other):
        """
        Return the union of two :class:`GenomicRange`
        """
        if (self.seq_id == other.seq_id) and (self.strand == other.strand):
            result = union_range(self.start, self.end, other.start, other.end)
            if result is not None:

                gen_range = GenomicRange()
                gen_range.seq_id = self.seq_id
                gen_range.strand = self.strand
                gen_range.start = result[0]
                gen_range.end = result[1]

                return gen_range

        return None

    def __or__(self, other):
        return self.union(other)

    def expand_from_list(self, others):
        """
        Expand the :class:`GenomicRange` range instance with a list of
        :class:`GenomicRange`

        Arguments:
            others (iterable): iterable of :class:`GenomicRange`
        """
        new_range = self

        for other in others:
            union = new_range.union(other)
            if union is None:
                continue
            new_range = union

        self.start = new_range.start
        self.end = new_range.end

    def intersect(self, other):
        """
        Return an instance of :class:`GenomicRange` that represent the
        intersection of the current instance and another.
        """
        if (self.seq_id == other.seq_id) and (self.strand == other.strand):

            if between(other.start, self.start, self.end) or \
               between(other.end, self.start, self.end) or \
               between(self.start, other.start, other.end) or \
               between(self.end, other.start, other.end):

                gen_range = GenomicRange()
                gen_range.start = max(self.start, other.start)
                gen_range.end = min(self.end, other.end)

                return gen_range

        return None

    def __and__(self, other):
        return self.intersect(other)

    def __contains__(self, pos):
        """

        .. versionchanged:: 0.2.3
            a range or a subclass are accepted

        .. versionadded:: 0.1.16

        Tests if the position is inside the range of the GenomicRange

        Pos is 1-based as :attr:`GenomicRange.start` and
        :attr:`GenomicRange.end`
        """
        if isinstance(pos, int):
            return between(pos, self.start, self.end)
        elif isinstance(pos, GenomicRange):
            pos = (pos.start, pos.end)
        return (between(pos[0], self.start, self.end)) and \
            (between(pos[1], self.start, self.end))

    def get_range(self):
        """
        .. versionadded:: 0.1.13

        Returns the start and end position as a tuple
        """
        return (self.start, self.end)

    def get_relative_pos(self, pos):
        """
        .. versionadded:: 0.1.16

        Given an absolute position (referred to the reference), convert the
        position to a coordinate relative to the GenomicRange

        Returns:
            int: the position relative to the GenomicRange

        Raises:
            ValueError: if the position is not in the range
        """
        if pos not in self:
            raise ValueError("Position {} not in GenomicRange".format(pos))
        return pos - self.start + 1


class Annotation(GenomicRange):
    """
    .. versionadded:: 0.1.12

    .. versionchanged:: 0.2.1
        using __slots__ for better memory usage

    Alternative implementation for an Annotation. When initialised, If *uid* is
    None, a unique id is added using `uuid.uuid4`.
    """

    __slots__ = ('source', 'feat_type', 'score', 'phase', 'attr')

    # source = 'None'
    # "Annotation source"
    # feat_type = 'None'
    # "Annotation type (e.g. CDS, gene, exon, etc.)"
    # score = 0.0
    # "Score associated to the annotation"
    # phase = 0
    # "Annotation phase, (0, 1, 2)"
    # attr = None
    # "Dictionary with the key value pairs in the last column of a GFF/GTF"

    def __init__(self, seq_id='None', start=1, end=1, strand='+',
                 source='None', feat_type='None', score=0.0, phase=0, uid=None,
                 **kwd):
        super(Annotation, self).__init__(
            seq_id=seq_id,
            start=start,
            end=end,
            strand=strand
        )

        self.source = source
        self.feat_type = feat_type
        self.score = score
        self.phase = phase

        self.attr = kwd

        if uid is None:
            self.uid = str(uuid.uuid4())
        else:
            self.uid = uid

    def __hash__(self):
        return hash(self.uid)

    def get_ec(self, level=4):
        """
        .. versionadded:: 0.1.13

        .. versionchanged:: 0.2.0
            returns a *set* instead of a list

        Returns the EC values associated with the annotation, cutting them at
        the desired level.

        Arguments:
            level (int): level of classification desired (between 1 and 4)

        Returns:
            set: list of all EC numbers associated, at the desired level, if
            none are found an empty set is returned
        """
        ec = self.attr.get('EC', None)
        if ec is None:
            return set()

        ec = ec.split(',')

        return set(['.'.join(x.split('.')[:level]) for x in ec])

    def get_mapping(self, db):
        """
        .. versionadded:: 0.1.13

        Returns the mappings, to a particular db, associated with the
        annotation.

        Arguments:
            db (str): database to which the mappings come from

        Returns:
            list: list of all mappings associated, to the specified db, if
            none are found an empty list is returned
        """
        mappings = self.attr.get('map_{0}'.format(db.upper()))
        if mappings is None:
            return []

        return mappings.split(',')

    def set_mapping(self, db, values):
        """
        .. versionadded:: 0.1.13

        Set mappings to a particular db, associated with the
        annotation.

        Arguments:
            db (str): database to which the mappings come from
            mappings (iterable): iterable of mappings

        """
        self.set_attr(
            'map_{0}'.format(db.upper()),
            ','.join(values)
        )

    def get_mappings(self):
        """
        .. versionadded:: 0.2.1

        Return a dictionary where the keys are the mapping DBs (lowercase) and
        and the values are the mapping IDs for that DB
        """
        mappings = {}

        for key in self.attr:
            if key.startswith('map_'):
                db = key.replace('map_', '').lower()
                mappings[db] = self.get_mapping(db)

        return mappings

    @property
    def taxon_id(self):
        """
        .. versionchanged:: 0.3.1
            if taxon_id is set to "None" as a string, it's converted to *None*

        taxon_id of the annotation
        """
        value = self.attr.get('taxon_id', None)
        try:
            value = int(value)
        except (TypeError, ValueError):
            value = None
        return value

    @taxon_id.setter
    def taxon_id(self, value):
        self.attr['taxon_id'] = int(value)

    @property
    def db(self):
        "db used for the gene_id prediction"
        return self.attr.get('db', None)

    @db.setter
    def db(self, value):
        self.attr['db'] = value

    @property
    def taxon_db(self):
        "db used for the taxon_id prediction"
        return self.attr.get('taxon_db', None)

    @taxon_db.setter
    def taxon_db(self, value):
        self.attr['taxon_db'] = value

    @property
    def dbq(self):
        "db quality of the annotation"
        try:
            return self.get_attr('dbq', int)
        except AttributeNotFound:
            return None

    @dbq.setter
    def dbq(self, value):
        self.attr['dbq'] = value

    @property
    def uid(self):
        """
        .. versionadded:: 0.1.13

        uid of the annotation
        """
        value = self.attr.get('uid', None)
        if value is None:
            # old data where the unique id is marked as ko_idx
            value = self.attr.get('ko_idx', None)

        return value

    @uid.setter
    def uid(self, value):
        self.attr['uid'] = value

    @property
    def bitscore(self):
        "bitscore of the annotation"
        try:
            return float(self.attr['bitscore'])
        except KeyError:
            # legacy for old data
            bitscore = self.attr.get('bit_score', None)
            return None if bitscore is None else float(bitscore)

    @bitscore.setter
    def bitscore(self, value):
        self.attr['bitscore'] = float(value)

    @property
    def gene_id(self):
        "gene_id of the annotation, or *ko* if available"
        try:
            return self.attr['gene_id']
        except KeyError:
            # legacy for old data
            return self.attr.get('ko', None)

    @gene_id.setter
    def gene_id(self, value):
        self.attr['gene_id'] = value

    @property
    def length(self):
        """
        .. versionchanged:: 0.2.0

        Length of the annotation, uses `len(self)`
        """
        return len(self)

    @property
    def region(self):
        """
        .. versionadded:: 0.1.13

        Return the *region* covered by the annotation, to use in samtools
        """
        return "{0}:{1}:{2}".format(self.seq_id, self.start, self.end)

    @property
    def counts(self):
        """
        .. versionadded:: 0.2.2

        Returns the sample counts for the annotation
        """
        counts = {}

        for key, value in viewitems(self.attr):
            if key.startswith('counts_'):
                key = key.replace('counts_', '')
                counts[key] = float(value)

        return counts

    @counts.setter
    def counts(self, counts):
        """
        .. versionadded:: 0.2.2

        Sets the sample counts for the annotation

        Arguments:
            counts (dict): key is the sample name and the count for it
        """
        for key, value in viewitems(counts):
            self.attr["counts_{}".format(key)] = value

    @property
    def fpkms(self):
        """
        .. versionadded:: 0.2.2

        Returns the sample fpkms for the annotation
        """
        fpkms = {}

        for key, value in viewitems(self.attr):
            if key.startswith('fpkms_'):
                key = key.replace('fpkms_', '')
                fpkms[key] = float(value)

        return fpkms

    @fpkms.setter
    def fpkms(self, fpkms):
        """
        .. versionadded:: 0.2.2

        Sets the sample fpkms for the annotation

        Arguments:
            fpkms (dict): key is the sample name and the fpmk for it
        """
        for key, value in viewitems(fpkms):
            self.attr["fpkms_{}".format(key)] = value

    def add_exp_syn_count(self, seq, syn_matrix=None):
        """
        .. versionadded:: 0.1.13

        Adds expected synonymous/non-synonymous values for an annotation.

        Arguments:
            seq (str): sequence corresponding to the annotation seq_id
                syn_matrix (None, dict): matrix that determines the return
                values. Defaults to the one defined in the called function
                :func:`mgkit.utils.sequnce.get_seq_expected_syn_count`.

        """

        seq = self.get_nuc_seq(seq, reverse=self.strand == '-')

        syn_count, nonsyn_count = seq_utils.get_seq_expected_syn_count(
            seq,
            syn_matrix=syn_matrix
        )

        self.set_attr('exp_syn', syn_count)
        self.set_attr('exp_nonsyn', nonsyn_count)

    def to_gff(self, sep='='):
        """
        Format the Annotation as a GFF string.

        Arguments:
            sep (str): separator key -> value

        Returns:
            str: annotation formatted as GFF
        """
        var_names = (
            'seq_id', 'source', 'feat_type', 'start', 'end',
            'score', 'strand', 'phase'
        )

        values = '\t'.join(
            str(getattr(self, var_name))
            for var_name in var_names
        )

        attr_column = ';'.join(
            '{0}{1}"{2}"'.format(
                key,
                sep,
                quote(str(self.attr[key]), ' ()/')
            )
            for key in sorted(self.attr)
        )

        return "{0}\t{1}\n".format(values, attr_column)

    def to_dict(self, exclude_attr=None):
        """
        .. versionadded:: 0.3.1

        Return a dictionary representation of the Annotation.

        Arguments:
            exclude_attr (str,list): attributes to exclude from the dictionary,
                can be either a single attribute (string) or a list of strings

        Returns:
            dict: dictionary with the annotation
        """
        var_names = (
            'seq_id', 'source', 'feat_type', 'start', 'end',
            'score', 'strand', 'phase'
        )

        dictionary = {}

        for var_name in var_names:
            dictionary[var_name] = getattr(self, var_name)

        dictionary.update(self.attr)

        if exclude_attr is not None:
            if isinstance(exclude_attr, str):
                exclude_attr = [exclude_attr]

            for attr in exclude_attr:
                del dictionary[attr]

        return dictionary

    def to_json(self):
        """
        .. versionadded:: 0.2.1

        .. versionchanged:: 0.3.1
            now :meth:`Annotation.to_dict` is used

        Returns a json representation of the Annotation
        """

        return json.dumps(self.to_dict(), separators=(',', ':'))

    def to_mongodb(self, lineage_func=None, indent=None, raw=False):
        """
        .. versionadded:: 0.2.1

        .. versionchanged:: 0.2.2
            added handling of *counts_* and *fpkms_*

        .. versionchanged:: 0.2.6
            added *indent* parameter

        .. versionchanged:: 0.3.4
            added *raw*

        Returns a MongoDB document that represent the Annotation.

        Arguments:
            lineage (func): function used to populate the lineage key, returns
                a list of taxon_id
            indent (int): the amount of indent to put in the record, None (the
                default) is for the most compact - one line for the record
            raw (bool): if True, the method returns a string, which is the
                json dump, if False, the value returned is the dictionary

        Returns:
            str or dict: the MongoDB document, with Annotation.uid as _id, as
            a string if *raw* is True, a dictionary if it is False
        """

        # OrderedDict is necessary to keep the order of the keys
        dictionary = OrderedDict()

        # _id must be the first element
        dictionary['_id'] = self.uid

        var_names = (
            'seq_id', 'source', 'feat_type', 'start', 'end', 'score', 'strand',
            'phase', 'gene_id', 'taxon_id', 'bitscore', 'exp_nonsyn',
            'exp_syn', 'length', 'dbq', 'coverage', 'uid'
        )

        for var_name in var_names:
            try:
                dictionary[var_name] = getattr(self, var_name)
            except AttributeNotFound:
                pass

        ec_ids = self.get_ec()
        mappings = self.get_mappings()

        # if one at least has values
        if ec_ids:
            mappings['ec'] = list(ec_ids)

        if lineage_func is not None:
            dictionary['lineage'] = lineage_func(self.taxon_id)

        dictionary['map'] = mappings

        counts = self.counts
        if counts:
            dictionary['counts'] = counts

        fpkms = self.fpkms
        if fpkms:
            dictionary['fpkms'] = fpkms

        # the rest of the dictionary should be put, excluding special keys:
        # uid is used as _id in the document
        # EC is put as a array, as is any mapping like map_KO
        dictionary.update(
            dict(
                (key, value)
                for key, value in viewitems(self.attr)
                if (key not in var_names) and (key not in ('uid', 'EC')) and
                (not key.startswith('map_')) and
                (not key.startswith('counts_')) and
                (not key.startswith('fpkms_'))
            )
        )

        if raw:
            return dictionary

        return json.dumps(dictionary, indent=indent, separators=(',', ':'))

    def to_file(self, file_handle):
        """
        Writes the GFF annotation to *file_handle*
        """
        file_handle.write(self.to_gff().encode('ascii'))

    def to_gtf(self, gene_id_attr='uid', sep=' '):
        """
        .. versionadded:: 0.1.15

        .. versionchanged:: 0.1.16
            added *gene_id_attr* parameter

        .. versionchanged:: 0.2.2
            added *sep* argument, default to a space, now

        Simple conversion to a valid GTF. gene_id and transcript_id are set to
        *uid* or the attribute specified using the *gene_id_attr* parameter.
        It's written to be used with *SNPDat*.
        """
        var_names = (
            'seq_id', 'source', 'feat_type', 'start', 'end',
            'score', 'strand', 'phase'
        )

        values = '\t'.join(
            str(getattr(self, var_name))
            for var_name in var_names
        )

        # Keys that needs to be at the start of the attributes
        gtf_attr = ['gene_id', 'transcript_id']

        attr_keys = sorted(self.attr.keys())

        # eliminate gene_id (always present in new ones)
        try:
            attr_keys.remove('gene_id')
        except ValueError:
            pass

        # transcript_id don't always be there
        try:
            attr_keys.remove('transcript_id')
        except ValueError:
            pass

        attr_values = [self.get_attr(gene_id_attr)] * 2 + [
            self.attr[attr_key]
            for attr_key in attr_keys
        ]
        attr_keys = gtf_attr + attr_keys

        attr_column = '; '.join(
            '{0}{1}"{2}"'.format(
                key,
                sep,
                quote(value, ' ()/')
            )
            for key, value in zip(attr_keys, attr_values)
        )

        return "{0}\t{1}\n".format(values, attr_column)

    @property
    def sample_coverage(self):
        """
        .. versionadded:: 0.1.13

        Returns a dictionary with the coverage for each sample, the returned
        dictionary has the sample id (stripped of the *_cov*) suffix and as
        values the coverage (converted via :func:`int`).

        :return dict: dictionary with the samples' coverage
        """
        attributes = self.attr

        return dict(
            (attribute.replace('_cov', ''), float(value))
            for attribute, value in viewitems(attributes)
            if attribute.endswith('_cov')
        )

    def get_number_of_samples(self, min_cov=MIN_COV):
        """
        .. versionadded:: 0.1.13

        Returns the number of sample that have at least a minimum coverage of
        `min_cov`.

        :param int min_cov: minimum coverage
        :return int: number of samples passing the filter
        :raise AttributeNotFound: if no sample coverage attribute is found
        """
        coverage = self.sample_coverage

        if not coverage:
            raise AttributeNotFound(
                'No coverage attribute found (ending in "_cov")'
            )

        return sum(
            1 for sample, coverage in viewitems(coverage)
            if coverage >= min_cov
        )

    def get_attr(self, attr, conv=str):
        """

        .. versionchanged:: 0.3.4
            any GFF attribute can be returned

        .. versionchanged:: 0.3.3
            added *seq_id* as special attribute, in addition do *length*

        .. versionadded:: 0.1.13

        Generic method to get an attribute and convert it to a specific
        datatype. The order for the lookup is:

        * length
        * self.attr (dictionary)
        * getattr(self) of the first 8 columns of a GFF (seq_id, source, ...)
        """
        if attr == 'length':
            return len(self)

        value = self.attr.get(attr, None)
        if value is not None:
            return conv(value)

        if attr in ('seq_id', 'source', 'feat_type', 'start', 'end', 'score',
                    'strand', 'phase'):
            value = getattr(self, attr, None)
            if value is not None:
                return conv(value)

        raise AttributeNotFound('No {0} attribute found'.format(attr))

    def set_attr(self, attr, value):
        """
        .. versionadded:: 0.1.13

        .. versionchanged:: 0.4.4
            a standard attribute can now be set

        Generic method to set an attribute
        """

        if attr in ('seq_id', 'source', 'feat_type', 'start', 'end', 'score',
                    'strand', 'phase'):
            setattr(self, attr, value)
            return None

        self.attr[attr] = value

    def has_attr(self, attr):
        """
        .. versionadded:: 0.4.4

        Tests if an attributes is present in the Annotation

        Arguments:
            attr (str): attribute to test

        Returns:
            bool: True if the attributes is present

        """
        if attr in ('seq_id', 'source', 'feat_type', 'start', 'end', 'score',
                    'strand', 'phase'):
            return True

        return attr in self.attr

    def del_attr(self, attr):
        """
        .. versionadded:: 0.4.4

        Removes attributes from the Annotation
        """

        if attr in ('seq_id', 'source', 'feat_type', 'start', 'end', 'score',
                    'strand', 'phase'):
            return None

        try:
            del self.attr[attr]
        except KeyError:
            pass

    @property
    def coverage(self):
        """
        .. versionadded:: 0.1.13

        Return the total coverage for the annotation

        :return float: coverage
        :raise AttributeNotFound: if no coverage attribute is found
        """
        return self.get_attr('cov', float)

    @property
    def exp_syn(self):
        """
        .. versionadded:: 0.1.13

        Returns the expected number of synonymous changes
        """
        return self.get_attr('exp_syn', int)

    @property
    def exp_nonsyn(self):
        """
        .. versionadded:: 0.1.13

        Returns the expected number of non-synonymous changes
        """
        return self.get_attr('exp_nonsyn', int)

    def get_nuc_seq(self, seq, reverse=False, snp=None):
        """
        .. versionadded:: 0.1.13

        .. versionchanged:: 0.1.16
            added *snp* parameter

        Returns the nucleotidic sequence that the annotation covers. if the
        annotation's strand is *-*, and *reverse* is True, the reverse
        complement is returned.

        Arguments:
            seq (seq): chromosome/contig sequence
            reverse (bool): if True and the strand is '-', a reverse complement
                is returned
            snp (tuple): first element is the position of the SNP relative to
                the Annotation and the second element is the change

        Returns:
            str: nucleotide sequence with requested transformations

        """
        ann_seq = seq[self.start - 1:self.end]
        if snp is not None:
            ann_seq = seq_utils.get_variant_sequence(ann_seq, snp)

        if (self.strand == '-') and reverse:
            ann_seq = seq_utils.reverse_complement(ann_seq)

        return ann_seq

    def get_aa_seq(self, seq, start=0, tbl=None, snp=None):
        """
        .. versionadded:: 0.1.16

        Returns a translated aminoacid sequence of the annotation. The snp
        parameter is passed to :meth:`Annotation.get_nuc_seq`

        Arguments:
            seq (seq): chromosome/contig sequence
            start (int): position (0-based) from where the correct occurs
                (frame). If None, the phase attribute is used
            tbl (dict): dictionary with the translation for each codon,
                passed to :func:`mgkit.utils.sequence.translate_sequence`
            snp (tuple): first element is the position of the SNP and the
                second element is the change

        Returns:
            str: aminoacid sequence
        """

        if start is None:
            start = self.phase

        nuc_seq = self.get_nuc_seq(seq, reverse=True, snp=snp)
        return seq_utils.translate_sequence(
            nuc_seq,
            start=start,
            tbl=None,
            reverse=False
        )

    def add_gc_content(self, seq):
        """
        Adds GC content information for an annotation. The formula is:

        .. math::
            :label: gc_content_gff

            \\frac {(G + C)}{(G + C + A + T)}

        Modifies the instances of the annotation. gc_ratio will be added to its
        attributes.

        Arguments:
            seq (str): nucleotide sequence referred in the GFF

        """

        ann_seq = self.get_nuc_seq(
            seq,
            reverse=True if self.strand == '-' else False
        )

        at_sum = (ann_seq.count('A') + ann_seq.count('T'))
        gc_sum = (ann_seq.count('G') + ann_seq.count('C'))

        gc_cont = gc_sum / (gc_sum + at_sum)

        self.set_attr('gc_cont', gc_cont)

    def add_gc_ratio(self, seq):
        """
        Adds GC content information for an annotation. The formula is:

        .. math::
            :label: gc_ratio_gff

            \\frac {(A + T)}{(G + C)}

        Modifies the instances of the annotation. gc_ratio will be added to its
        attributes.

        Arguments:
            seq (str): nucleotide sequence referred in the GFF

        """

        ann_seq = self.get_nuc_seq(
            seq,
            reverse=True if self.strand == '-' else False
        )

        at_sum = (ann_seq.count('A') + ann_seq.count('T'))
        gc_sum = (ann_seq.count('G') + ann_seq.count('C'))

        gc_ratio = at_sum / gc_sum

        self.set_attr('gc_ratio', gc_ratio)

    def is_syn(self, seq, pos, change, tbl=None, abs_pos=True, start=0,
               strict=True):
        """
        .. versionadded:: 0.1.16

        .. versionchanged:: 0.4.4
            added *strict* parameter

        Return if a SNP is synonymous or non-synonymous.

        Arguments:
            seq (seq): reference sequence of the annotation
            pos (int): position of the SNP on the reference (1-based index)
            change (str): nucleotidic change
            tbl (dict): dictionary with the translation table. Defaults to the
                universal genetic code
            abs_pos (bool): if True the *pos* is referred to the reference and
                not a position relative to the annotation
            start (int or None): phase to be used to get the start position of
                the codon. if None, the Annotation phase will be used
            strict (bool): if a variant codon is not found, a KeyError is
                raised, otherwise *False* is returned

        Returns:
            bool: True if the SNP is synonymous, false if it's non-synonymous.
            Behaviour in case of variant codons not found in the translation
            table changes based on *strict*

        Raises:
            KeyError: if the variant codon is not found and *strict* is True
        """
        if abs_pos:
            rel_pos = self.get_relative_pos(pos)
        else:
            rel_pos = pos

        if start is None:
            start = self.phase

        if tbl is None:
            tbl = UNIVERSAL

        # codon number in the sequence
        codon_index = (rel_pos - start - 1) // 3
        # the position to slice the seq to get a codon (0-based). It takes into
        # account the phase (start) and the codon index
        seq_start = (self.start + start + (codon_index * 3) - 1)
        # position in the codon using the relative position and the phase/frame
        # the module will give 1, 2 or 0. -1 will shift the position correctly
        codon_change = ((rel_pos - start) % 3) - 1

        codon = seq[seq_start:seq_start+3]

        var_codon = list(codon)
        var_codon[codon_change] = change
        var_codon = ''.join(var_codon)

        if self.strand == '-':
            codon = seq_utils.reverse_complement(codon)
            var_codon = seq_utils.reverse_complement(var_codon)

        try:
            return UNIVERSAL[codon] == UNIVERSAL[var_codon]
        except KeyError:
            LOG.warning("""Annotation %s has an unrecognised codon:
                        reference %s, variant %s""", self.uid, codon,
                        var_codon)
            if strict:
                raise KeyError("Variant codon not found %s", var_codon)
            else:
                return False


def from_glimmer3(header, line, feat_type='CDS'):
    """
    .. versionadded:: 0.1.12

    Parses the line of a GLIMMER3 ouput and returns an instance of a GFF
    annotation.

    Arguments:
        header (str): the seq_id to which the ORF belongs
        line (str): the prediction line for the orf
        feat_type (str): the feature type to use

    Returns:
        Annotation: instance of annotation

    Example:
        Assuming a GLIMMER3 output like this::

            >sequence0001
            orf00001       66      611  +3     6.08

        The code used is:

        >>> header = 'sequence0001'
        >>> line = 'orf00001       66      611  +3     6.08'
        >>> from_glimmer3(header, line)

    """
    if isinstance(line, bytes):
        line = line.decode('ascii')
    orf_id, start, end, frame, score = line.split()

    start = int(start)
    end = int(end)

    if start > end:
        start, end = end, start

    annotation = Annotation(
        seq_id=header,
        source='GLIMMER3',
        feat_type=feat_type,
        start=start,
        end=end,
        score=float(score),
        strand=frame[0],
        phase=int(frame[1]) - 1,
        frame=frame,
        glimmer_score=float(score),
        orf_id=orf_id
    )

    return annotation


class DuplicateKeyError(Exception):
    """
    .. versionadded:: 0.1.12

    Raised if a GFF annotation contains duplicate keys
    """
    pass


def from_gff(line, strict=True, encoding='ascii'):
    """
    .. versionadded:: 0.1.12

    .. versionchanged:: 0.2.6
        added *strict* parameter

    .. versionchanged:: 0.4.0
        added *encoding* parameter

    Parse GFF line and returns an :class:`Annotation` instance

    Arguments:
        line (str): GFF line
        strict (bool): if True duplicate keys raise an exception

    Returns:
        Annotation: instance of :class:`Annotation` for the line

    Raises:
        DuplicateKeyError: if the attribute column has duplicate keys

    """
    if isinstance(line, bytes):
        line = line.decode(encoding)
    line = line.rstrip().split('\t')

    # in case the last column (attributes) is empty
    if len(line) < 9:
        values = line
        # bug in which the phase was not written
        if len(line[-1]) > 1:
            line.insert(-1, 0)
    else:
        values = line[:-1]

    var_names = (
        'seq_id', 'source', 'feat_type', 'start', 'end',
        'score', 'strand', 'phase'
    )
    # the phase sometimes can be set as unknown, using '-'. We prefer using 0
    var_types = (str, str, str, int, int, float, str,
                 lambda x: 0 if x == '' else int(x))

    attr = {}

    for var, value, vtype in zip(var_names, values, var_types):
        try:
            attr[var] = vtype(value)
        except ValueError:
            attr[var] = value

    # in case the last column (attributes) is empty
    if len(line) < 9:
        return Annotation(**attr)

    for pair in line[-1].split(';'):
        try:
            # by default the key,value separator '=' is assumed to be used
            var, value = pair.strip().split('=', 1)
        except ValueError:
            # in case it doesn't work, it is assumed to be a space
            if ' ' in pair.strip():
                var, value = pair.strip().split(' ', 1)
            else:
                # case in which there's an attribute but no value, like a bool
                var = pair.strip()
                value = None

        if (var in attr) and strict:
            raise DuplicateKeyError("Duplicate attribute: {0}".format(var))

        # skips possible key/values generated by the line ending with a ';'
        if not var:
            continue

        if value is not None:
            value = unquote(value.replace('"', ''))
        attr[var] = value

    return Annotation(**attr)


def from_sequence(name, seq, feat_type='SEQUENCE', **kwd):
    """
    .. versionadded:: 0.1.12

    Returns an instance of :class:`Annotation` for the full length of a
    sequence

    Arguments:
        name (str): name of the sequence
        seq (str): sequence, to get the length of the annotation

    Keyword Args:
        feat_type (str): feature type in the GFF
        **kwd: any additional column

    Returns:
        Annotation: instance of :class:`Annotation`

    """
    annotation = Annotation(
        seq_id=name,
        source='SEQUENCE',
        feat_type=feat_type,
        start=1,
        end=len(seq),
        score=0.0,
        strand='+',
        phase=0,
        sequence=name,
        **kwd
    )

    return annotation


def from_aa_blast_frag(hit, parent_ann, aa_seqs):
    frag_id, frame = hit[0].split('-')
    strand = '+' if frame.startswith('f') else '-'
    frame = int(frame[1])
    identity = hit[2]
    bitscore = hit[-1]
    start = hit[3]
    end = hit[4]
    if strand == '-':
        start, end = seq_utils.reverse_aa_coord(
            start,
            end,
            len(aa_seqs[hit[0]])
        )
    start, end = seq_utils.convert_aa_to_nuc_coord(start, end, frame)

    annotation = Annotation(
        seq_id=parent_ann.seq_id,
        source='BLAST',
        feat_type='CDS',
        start=start + parent_ann.start - 1,
        end=end + parent_ann.start - 1,
        score=bitscore,
        strand=strand,
        phase=frame,
        db='UNIPROT',
        gene_id=hit[1],
        identity=identity,
        bitscore=bitscore,
        ID=frag_id
    )

    return annotation


def from_nuc_blast_frag(hit, parent_ann, db='NCBI-NT'):
    frag_id = hit[0]
    strand = '+'
    identity = hit[2]
    bitscore = hit[-1]
    start = hit[3]
    end = hit[4]

    annotation = Annotation(
        seq_id=parent_ann.seq_id,
        source='BLAST',
        feat_type='CDS',
        start=start + parent_ann.start - 1,
        end=end + parent_ann.start - 1,
        score=bitscore,
        strand=strand,
        phase=0,
        db=db,
        gene_id=hit[1],
        identity=identity,
        bitscore=bitscore,
        ID=frag_id
    )

    return annotation


def annotate_sequence(name, seq, window=None):

    length = len(seq)

    if window is None:
        window = length

    for index in range(1, length, window):
        annotation = Annotation.from_sequence(name, seq)
        annotation.start = index
        annotation.end = index + window - 1
        if annotation.end > length:
            annotation.end = length
        yield annotation


def from_nuc_blast(hit, db, feat_type='CDS', seq_len=None, to_nuc=False,
                   **kwd):
    """
    .. versionadded:: 0.1.12

    .. versionchanged:: 0.1.16
        added *to_nuc* parameter

    .. versionchanged:: 0.2.3
        removed *to_nuc*, the hit can include the subject end/start and evalue

    Returns an instance of :class:`Annotation`

    Arguments:
        hit (tuple): a BLAST hit, from :func:`mgkit.io.blast.parse_blast_tab`
        db (str): db used with BLAST

    Keyword Args:
        feat_type (str): feature type in the GFF
        seq_len (int): sequence length, if supplied, the phase for strand '-'
            can be assigned, otherwise is assigned a 0
        **kwd: any additional column

    Returns:
        Annotation: instance of :class:`Annotation`
    """
    seq_id = hit[0]
    gene_id = hit[1]
    strand = '+'
    identity = hit[2]
    bitscore = hit[-1]
    start = hit[3]
    end = hit[4]
    phase = 0

    if start > end:
        start, end = end, start
        strand = '-'
        if seq_len is not None:
            phase = (seq_len - end) % 3

    if strand == '+':
        phase = (start - 1) % 3

    annotation = Annotation(
        seq_id=seq_id,
        source='BLAST',
        feat_type=feat_type,
        start=start,
        end=end,
        score=bitscore,
        strand=strand,
        phase=phase,
        db=db,
        gene_id=gene_id,
        identity=identity,
        bitscore=bitscore,
        frame="{}{}".format('f' if strand == '+' else 'r', phase),
        **kwd
    )

    # the hit includes subject end/start and evalue, as per new version of
    # mgkit.io.blast.parse_uniprot_blast
    if len(hit) == 9:
        annotation.attr['evalue'] = hit[-2]
        annotation.attr['subject_end'] = hit[-3]
        annotation.attr['subject_start'] = hit[-4]

    return annotation


def from_json(line):
    """
        .. versionadded:: 0.2.1

        Returns an Annotation from a json representation
    """
    return Annotation(**json.loads(line))


def from_hmmer(line, aa_seqs, feat_type='gene', source='HMMER',
               db='CUSTOM', custom_profiles=True, noframe=False):
    """
    .. versionadded:: 0.1.15
        first implementation to move old scripts to new GFF specs

    .. versionchanged:: 0.2.1
        removed compatibility with old scripts

    .. versionchanged:: 0.2.2
        taxon_id and taxon_name are not saved for non-custom profiles

    .. versionchanged:: 0.3.1
        added support for non mgkit-translated sequences (*noframe*)

    Parse HMMER results (one line), it won't parse commented lines (starting
    with *#*)

    Arguments:
        line (str): HMMER domain table line
        aa_seqs (dict): dictionary with amino-acid sequences (name->seq),
            used to get the correct nucleotide positions
        feat_type (str): string to be used in the 'feature type' column
        source (str): string to be used in the 'source' column
        custom_profiles (bool): if True, the profile name contains gene,
            taxonomy and reviewed information in the form
            KOID_TAXONID_TAXON-NAME(-nr)
        noframe (bool): if True, the sequence is assumed to be in frame f0

    Returns:
        A :class:`Annotation` instance

    .. note::

        if `custom_profiles` is False, gene_id, taxon_id and taxon_name will
        be equal to the profile name

    """
    if isinstance(line, bytes):
        line = line.decode('ascii')
    line = line.split()
    if noframe:
        # no information on the frame is provided (already a protein, so f0)
        frame = 'f0'
        contig = line[0]
    else:
        contig, frame = line[0].rsplit('-', 1)

    t_from = int(line[17])
    t_to = int(line[18])
    # first get coordinate if sequence is reversed
    if frame.startswith('r'):
        seq_len = len(aa_seqs[line[0]])
        t_from, t_to = seq_utils.reverse_aa_coord(t_from, t_to, seq_len)
    # necessary only if frame information available
    if not noframe:
        # converts in nucleotide coordinates
        t_from, t_to = seq_utils.convert_aa_to_nuc_coord(
            t_from,
            t_to,
            frame=int(frame[-1])
        )

    # maintains the aa coordinates
    aa_from = int(line[17])
    aa_to = int(line[18])
    profile_name = line[3]
    score = float(line[6])

    if custom_profiles:
        # KOID_TAXONID_TAXON-NAME(-nr)
        reviewed = 'False' if profile_name.endswith('-nr') else 'True'
        gene_id, taxon_id, taxon_name = profile_name.split('_')
    else:
        gene_id = profile_name
        taxon_id = taxon_name = None

    annotation = Annotation(
        seq_id=contig,
        source=source,
        feat_type=feat_type,
        start=t_from,
        end=t_to,
        score=score,
        strand='-' if frame.startswith('r') else '+',
        phase=int(frame[1]),
        db=db,
        gene_id=gene_id,
        taxon_id=taxon_id,
        bitscore=float(line[7]),

        # custom for HMMER profiles
        aa_from=aa_from,
        aa_to=aa_to,
        # stores the aa sequence
        aa_seq=aa_seqs[line[0]][aa_from - 1:aa_to],
        # evalue
        evalue=score,

        # maintains HMMER profile information:
        # profile name
        name=profile_name,
        # both strand/phase (e.g r2)
        frame=frame,
        # old version of uid
        # ko_idx=ko_idx,
        # used in other old profiles, where the taxon name was used instead
        # of a taxon ID
        taxon_name=taxon_name
    )
    try:
        annotation.attr['reviewed'] = reviewed
    except UnboundLocalError:
        pass

    # removes the None values from non-custom profiles
    if taxon_id is None:
        del annotation.attr['taxon_id']

    if taxon_name is None:
        del annotation.attr['taxon_name']

    return annotation


def parse_gff(file_handle, gff_type=from_gff, strict=True, encoding='ascii'):
    """
    .. versionchanged:: 0.4.0
        In some cases ASCII decoding is not enough, so it is parametrised now

    .. versionchanged:: 0.3.4
        added decoding from binary for compatibility with Python3

    .. versionchanged:: 0.2.6
        added *strict* parameter

    .. versionchanged:: 0.2.3
        correctly handling of GFF with comments of appended sequences

    .. versionchanged:: 0.1.12
        added *gff_type* parameter

    Parse a GFF file and returns generator of :class:`GFFKegg` instances

    Accepts a file handle or a string with the file name

    Arguments:
        file_handle (str, file): file name or file handle to read from
        gff_type (class): class/function used to parse a GFF annotation
        strict (bool): if True duplicate keys raise an exception
        encoding (str): encoding of the file, if ascii fails, use utf8

    Yields:
        Annotation: an iterator of :class:`Annotation` instances
    """

    file_handle = mgkit.io.open_file(file_handle, 'rb')

    LOG.info(
        "Loading GFF from file (%s)",
        getattr(file_handle, 'name', repr(file_handle))
    )

    index = 0

    for index, line in enumerate(file_handle):
        try:
            line = line.decode(encoding)
        except UnicodeError:
            raise UnicodeError("Impossible to decode line to {}: {}".format(encoding, line))
        # the first is for GFF with comments and the second for
        # GFF with the fasta file attached
        if line.startswith('#'):
            continue
        if line.startswith('>'):
            break

        annotation = gff_type(line, strict=strict)
        yield annotation

    LOG.info(
        "Read %d line from file (%s)",
        index + 1,
        getattr(file_handle, 'name', repr(file_handle))
    )


def diff_gff(files, key_func=None):
    """
    .. versionadded:: 0.1.12

    Returns a simple diff made between a list of gff files. The annotations are
    grouped using *key_func*, so it depends on it to find similar annotations.

    Arguments:
        files (iterable): an iterable of file handles, pointing to GFF files
        key_func (func): function used to group annotations, defaults to this
            key: *(x.seq_id, x.strand, x.start, x.end, x.gene_id, x.bitscore)*

    Returns:
        dict: the returned dictionary keys are determined by key_func and as
        values lists. The lists elements are tuple whose first element is the
        index of the file, relative to *files* and the second element is the
        line number in which the annotation is. Can be used with the
        :mod:`linecache` module.
    """
    if isinstance(files, str) or len(files) == 1:
        return

    if key_func is None:
        def key_func(x):
            return (x.seq_id, x.strand, x.start, x.end, x.gene_id, x.bitscore)

    gff_diff = {}

    for index, file_handle in enumerate(files):
        for lineno, annotation in enumerate(parse_gff(file_handle)):
            key = key_func(annotation)
            try:
                gff_diff[key].append((index, lineno))
            except KeyError:
                gff_diff[key] = [(index, lineno)]

    return gff_diff


def annotation_elongation(ann1, annotations):
    """
    .. versionadded:: 0.1.12

    Given an :class:`Annotation` instance and a list of the instances of the
    same class, returns the longest overlapping range that can be found and the
    annotations that are included in it.

    .. warning::

        annotations are not checked for seq_id and strand

    Arguments:
        ann1 (Annotation): annotation to elongate
        annotations (iterable): iterable of :class:`Annotation` instances

    Returns:
        tuple: the first element is the longest range found, while the the
        second element is a set with the annotations used

    """
    used = set([ann1])

    union = (ann1.start, ann1.end)

    for ann2 in annotations:
        new_union = union_range(union[0], union[1], ann2.start, ann2.end)
        if new_union is not None:
            used.add(ann2)
            union = new_union

    return union, used


def elongate_annotations(annotations):
    """
    .. versionadded:: 0.1.12

    Given an iterable of :class:`Annotation` instances, tries to find the all
    possible longest ranges and returns them.

    .. warning::

        annotations are not checked for seq_id and strand

    Arguments:
        annotations (iterable): iterable of :class:`Annotation` instances

    Returns:
        set: set with the all ranges found
    """

    annotations = sorted(annotations, key=lambda x: x.start)

    ranges = set()

    while len(annotations) > 0:
        ann1 = annotations.pop(0)
        union, used = annotation_elongation(ann1, annotations)
        if union is None:
            ranges.add((ann1.start, ann1.end))
        else:
            annotations = sorted(set(annotations) - used,
                                 key=lambda x: x.start)
            ranges.add(union)

    return ranges


def annotation_coverage(annotations, seqs, strand=True):
    """
    .. versionadded:: 0.1.12

    Given a list of annotations and a dictionary where the keys are the
    sequence names referred in the annotations and the values are the sequences
    themselves, returns a number which indicated how much the sequence length
    is "covered" in annotations. If *strand* is True the coverage is strand
    specific.

    Arguments:
        annotations (iterable): iterable of :class:`Annotation` instances
        seqs (dict): dictionary in which the keys are the sequence names and
            the values are the sequences
        strand (bool): if True, the values are strand specific (the
            annotations) are grouped by (seq_id, strand) instead of seq_id

    Yields:
        tuple: the first element is the key, (seq_id, strand) if *strand* is
        True or seq_id if *strand* is False, and the coverage is the second
        value.
    """

    if strand:
        def key_func(x): return (x.seq_id, x.strand)
    else:
        def key_func(x): return x.seq_id

    annotations = group_annotations(
        annotations,
        key_func=key_func
    )

    for key, key_ann in viewitems(annotations):
        if isinstance(key, str):
            seq_len = len(seqs[key])
        else:
            seq_len = len(seqs[key[0]])

        covered = ranges_length(elongate_annotations(key_ann))

        yield key, covered / seq_len * 100


def annotation_coverage_sorted(annotations, seqs, strand=True):
    """
    .. versionadded:: 0.3.1

    Given a list of annotations and a dictionary where the keys are the
    sequence names referred in the annotations and the values are the sequences
    themselves, returns a number which indicated how much the sequence length
    is "covered" in annotations. If *strand* is True the coverage is strand
    specific.

    .. note::

        It differs from :func:`annotation_coverage` because it assumes the
        annotations are correctly sorted and in the values yielded

    Arguments:
        annotations (iterable): iterable of :class:`Annotation` instances
        seqs (dict): dictionary in which the keys are the sequence names and
            the values are the sequences
        strand (bool): if True, the values are strand specific (the
            annotations) are grouped by (seq_id, strand) instead of seq_id

    Yields:
        tuple: the first element is the seq_id, the second the strand (if
        strand is True, else it's set to *None*), and the third element is the
        coverage.
    """

    if strand:
        def key_func(x): return (x.seq_id, x.strand)
    else:
        def key_func(x): return x.seq_id

    annotations = group_annotations_sorted(
        annotations,
        key_func=key_func
    )

    for ann in annotations:
        seq_id = ann[0].seq_id
        if strand:
            ann_strand = ann[0].strand
        else:
            ann_strand = None
        seq_len = len(seqs[seq_id])

        covered = ranges_length(elongate_annotations(ann))

        yield seq_id, ann_strand, covered / seq_len * 100


def group_annotations(annotations, key_func=lambda x: (x.seq_id, x.strand)):
    """
    .. versionadded:: 0.1.12

    Group :class:`Annotation` instances in a dictionary by using a key function
    that returns the key to be used in the dictionary.

    Arguments:
        annotations (iterable): iterable with :class:`Annotation` instances
        key_func (func): function used to extract the key used in the
            dictionary, defaults to a function that returns
            (ann.seq_id, ann.strand)

    Returns:
        dict: dictionary whose keys are returned by *key_func* and the values
        are lists of annotations

    Example:
        >>> ann = [Annotation(seq_id='seq1', strand='+', start=10, end=15),
        ... Annotation(seq_id='seq1', strand='+', start=1, end=5),
        ... Annotation(seq_id='seq1', strand='-', start=30, end=100)]
        >>> group_annotations(ann)
        {('seq1', '+'): [seq1(+):10-15, seq1(+):1-5], ('seq1', '-'): [seq1(-):30-100]}
    """
    grouped = {}

    for annotation in annotations:
        key = key_func(annotation)
        try:
            grouped[key].append(annotation)
        except KeyError:
            grouped[key] = [annotation]

    return grouped


def group_annotations_sorted(annotations,
                             key_func=lambda x: (x.seq_id, x.strand)):
    """
    .. versionadded:: 0.1.13

    Group :class:`Annotation` instances by using a key function that returns a
    key. Assumes that the annotations are already sorted to return an iterator
    and save memory. One way to sort them is using: `sort -s -k 1,1 -k 7,7` on
    the file.

    Arguments:
        annotations (iterable): iterable with :class:`Annotation` instances
        key_func (func): function used to extract the key used in the
            dictionary, defaults to a function that returns
            (ann.seq_id, ann.strand)

    Yields:
        list: a list of the grouped annotations by *key_func* values

    """
    curr_key = ''
    curr_ann = []

    for annotation in annotations:
        new_key = key_func(annotation)
        if curr_key == new_key:
            curr_ann.append(annotation)
        else:
            if curr_key == '':
                curr_ann.append(annotation)
                curr_key = new_key
            else:
                yield curr_ann
                curr_key = new_key
                curr_ann = [annotation]
    else:
        yield curr_ann


def extract_nuc_seqs(annotations, seqs, name_func=lambda x: x.uid,
                     reverse=False):
    """
    .. versionadded:: 0.1.13

    Extract the nucleotidic sequences from a list of annotations. Internally
    uses the method :meth:`Annotation.get_nuc_seq`.

    Arguments:
        annotations (iterable): iterable of :class:`Annotation` instances
        seqs (dict): dictionary with the sequences referenced in the
            annotations
        name_func (func): function used to extract the sequence name to be
            used, defaults to the uid of the annotation
        reverse (bool): if True the annotations on the *-* strand are reverse
            complemented

    Yields:
        tuple: tuple whose first element is the sequence name and the second is
        the sequence to which the annotation refers.
    """
    for annotation in annotations:
        name = name_func(annotation)
        seq = annotation.get_nuc_seq(seqs[annotation.seq_id], reverse=reverse)

        yield name, seq


def group_annotations_by_ancestor(annotations, ancestors, taxonomy):
    """
    .. versionadded:: 0.1.13

    Group annotations by the ancestors provided.

    Arguments:
        annotations (iterable): annotations to group
        ancestors (iterable): list of ancestors accepted
        taxonomy: taxonomy class

    Returns:
        dict: grouped annotations
    """
    ann_dict = dict((ancestor, []) for ancestor in ancestors)

    unknown = []

    for annotation in annotations:
        anc_found = False
        for ancestor, anc_ids in viewitems(ancestors):
            if taxonomy.is_ancestor(annotation.taxon_id, anc_ids):
                ann_dict[ancestor].append(annotation)
                anc_found = True
                break
        if not anc_found:
            unknown.append(annotation)

    return ann_dict, unknown


def split_gff_file(file_handle, name_mask, num_files=2, encoding='ascii'):
    """
    .. versionadded:: 0.1.14

    .. versionchanged:: 0.2.6
        now accept a file object as sole input

    .. versionchanged:: 0.4.0
        added *encoding* parameter

    Splits a GFF, or a list of them, into a number of files. It is assured that
    annotations for the same sequence are kept in the same file, which is
    useful for cases like filtering, even when the annotations are from
    different GFF files.

    Internally, a structure is kept to check if a sequence ID is already been
    stored to a file, in which case the annotation is written to that file,
    otherwise a random file handles (among the open ones) is chosen.

    Arguments:
        file_handle (str, list): a single or list of file handles (or file
           names), from which the GFF annotations are read
        name_mask (str): a string used as template for the output file names
            on which the function applies :func:`string.format`
        num_files (int): the number of files to split the records

    Example:
        >>> import glob
        >>> files = glob.glob('*.gff')
        >>> name_mask = 'split-file-{0}.gff'
        >>> split_gff_file(files, name_mask, 5)
    """
    if not isinstance(file_handle, IOBase):
        if isinstance(file_handle, str):
            file_handle = [file_handle]

        file_handle = itertools.chain(
            *(mgkit.io.open_file(x, 'rb') for x in file_handle)
        )

    out_handles = [
        mgkit.io.open_file(name_mask.format(filen), 'wb')
        for filen in range(num_files)
    ]

    seq_ids = {}

    for line in file_handle:
        line = line.decode(encoding)
        if line.startswith('#'):
            continue
        if line.startswith('>'):
            break

        seq_id = line.split('\t')[0]
        try:
            out_handle = out_handles[seq_ids[seq_id]]
        except KeyError:
            new_index = random.randint(0, num_files - 1)
            seq_ids[seq_id] = new_index
            out_handle = out_handles[new_index]

        out_handle.write(line.encode('ascii'))


def load_gff_base_info(files, taxonomy=None, exclude_ids=None,
                       include_taxa=None, encoding='ascii'):
    """
    This function is useful if the number of annotations in a GFF is high or
    there are memory constraints on the system. It returns a dictionary that
    can be used with functions like
    :func:`mgkit.counts.func.load_sample_counts`.

    Arguments:
        files (iterable, str): file name or list of paths of GFF files
        taxonomy: taxonomy pickle file, needed if include_taxa is not None
        exclude_ids (set, list): a list of gene_id to exclude from the
            dictionary
        include_taxa (int, iterable): a taxon_id or list thereof to be passed
            to :meth:`mgkit.taxon.taxonomy.is_ancestor`, so only the taxa that
            have the those taxon_id(s) as ancestor(s) are kept
        encoding (str): passed to :func:`parse_gff`

    Returns:
        dict: dictionary where the key is :attr:`Annotation.uid` and the value
        is a tuple (:attr:`Annotation.gene_id`, :attr:`Annotation.taxon_id`)

    """
    if isinstance(files, str):
        files = [files]

    infos = {}

    for fname in files:
        for annotation in parse_gff(fname, encoding=encoding):
            # no information on taxa - exclude
            if annotation.taxon_id is None:
                continue
            # to exclude ribosomial genes or any other kind
            if exclude_ids is not None:
                if annotation.gene_id in exclude_ids:
                    continue
            if (include_taxa is not None) and (taxonomy is not None):
                if not taxonomy.is_ancestor(annotation.taxon_id, include_taxa):
                    continue

            infos[annotation.uid] = (annotation.gene_id, annotation.taxon_id)

    return infos


def load_gff_mappings(files, map_db, taxonomy=None, exclude_ids=None,
                      include_taxa=None, encoding='ascii'):
    """
    This function is useful if the number of annotations in a GFF is high or
    there are memory constraints on the system. It returns a dictionary that
    can be used with functions like
    :func:`mgkit.counts.func.load_sample_counts`.

    Arguments:
        files (iterable, str): file name or list of paths of GFF files
        map_db (str): any kind mapping in the GFF, as passed to
            :meth:`Annotation.get_mapping`
        taxonomy: taxonomy pickle file, needed if include_taxa is not None
        exclude_ids (set, list): a list of gene_id to exclude from the
            dictionary
        include_taxa (int, iterable): a taxon_id or list thereof to be passed
            to :meth:`mgkit.taxon.taxonomy.is_ancestor`, so only the taxa that
            have the those taxon_id(s) as ancestor(s) are kept
        encoding (str): passed to :func:`parse_gff`

    Returns:
        dict: dictionary where the key is :attr:`Annotation.gene_id` and the
        value is a list of mappings, as returned by
        :meth:`Annotation.get_mapping`

    """
    infos = {}

    for fname in files:
        for annotation in parse_gff(fname, encoding=encoding):
            # skips genes that are already in the mapping
            if annotation.gene_id in infos:
                continue
            # exclude genes with no taxonomic information
            if annotation.taxon_id is None:
                continue

            if exclude_ids is not None:
                if annotation.gene_id in exclude_ids:
                    continue

            # skips non bacterial/achaeal genes
            if (include_taxa is not None) and (taxonomy is not None):
                if not taxonomy.is_ancestor(annotation.taxon_id, include_taxa):
                    continue

            infos[annotation.gene_id] = annotation.get_mapping(map_db)

    return infos


def parse_gff_files(files, strict=True):
    """
    .. versionadded:: 0.1.15

    .. versionchanged:: 0.2.6
        added *strict* parameter

    Function that returns an iterator of annotations from multiple GFF files.

    Arguments:
        files (iterable, str): iterable of file names of GFF files, or a single
            file name
        strict (bool): if True duplicate keys raise an exception

    Yields:
        :class:`Annotation`: iterator of annotations
    """
    if isinstance(files, str):
        files = [files]

    return itertools.chain(*(parse_gff(file_name, strict=strict) for file_name in files))


def get_annotation_map(annotations, key_func, value_func):
    """
    .. versionadded:: 0.1.15

    Applies two functions to an iterable of annotations with an iterator
    returned with the applied functions. Useful to build a dictionary

    Arguments:
        annotations (iterable): iterable of annotations
        key_func (func): function that accept an annotation as argument and
            returns one value, the first of the returned tuple
        value_func (func): function that accept an annotation as argument and
            returns one value, the second of the returned tuple

    Yields:
        tuple: a tuple where the first value is the result of *key_func* on
        the passed annotation and the second is the value returned by
        *value_func* on the same annotation
    """
    for annotation in annotations:
        yield key_func(annotation), value_func(annotation)


def convert_gff_to_gtf(file_in, file_out, gene_id_attr='uid'):
    """
    .. versionadded:: 0.1.16

    Function that uses :meth:`Annotation.to_gtf` to convert a GFF into GTF.

    Arguments:
        file_in (str, file): either file name or file handle of a GFF file
        file_out (str): file name to which write the converted annotations
    """
    LOG.info("Writing GTF file to %s", file_out)
    file_out = open(file_out, 'w')
    for annotation in parse_gff(file_in):
        file_out.write(annotation.to_gtf())


def from_mongodb(record, lineage=True):
    """
    .. versionadded:: 0.2.1

    .. versionchanged:: 0.2.2
        added handling of *counts_* and *fpkms_*

    .. versionchanged:: 0.2.6
        better handling of missing attributes and added *lineage* parameter

    Returns a :class:`Annotation` instance from a MongoDB record (created)
    using :meth:`Annotation.to_mongodb`. The actual record returned by pymongo
    is a dictionary that is copied, manipulated and passed to the
    :meth:`Annotation.__init__`.

    Arguments:
        record (dict): a dictionary with the full record from a MongoDB query
        lineage (bool): indicates if the lineage information in the record
            should be kept in the annotation

    Returns:
        Annotation: instance of :class:`Annotation` object
    """
    record = record.copy()

    record['uid']
    del record['_id']

    if 'map' in record:

        mappings = record['map'].copy()
        del record['map']

        try:
            record['EC'] = ','.join(mappings['ec'])
            del mappings['ec']
        except KeyError:
            pass

        try:
            for key in mappings:
                record['map_{}'.format(key.upper())] = ','.join(mappings[key])
        except KeyError:
            pass

        try:
            counts = record['counts'].copy()
            del record['counts']
            for key in counts:
                record['counts_{}'.format(key)] = counts[key]
        except KeyError:
            pass

    try:
        fpkms = record['fpkms'].copy()
        del record['fpkms']
        for key in fpkms:
            record['fpkms_{}'.format(key)] = fpkms[key]
    except KeyError:
        pass

    if ('lineage' in record) and (not lineage):
        del record['lineage']

    return Annotation(**record)


def from_prodigal_frag(main_gff, blast_gff, attr='ID', split_func=None):
    """

    .. versionchanged:: 0.3.3
        fixed a bug for the strand, also the code is tested

    .. versionadded:: 0.2.6
        *experimental*

    Reads the GFF given in output by PRODIGAL and the resulting GFF from using
    BLAST (or other software) on the aa or nucleotide file output by PRODIGAL.

    It then integrates the two outputs, so to the PRODIGAL GFF is added the
    information from the the output of the gene prediction software used.

    Arguments:
        main_gff (file): GFF file from PRODIGAL
        blast_gff (file): GFF with the returned annotations
        attr (str): attribute in the PRODIGAL GFF that is used to identify an
            annotation
        split_func (func): function to rename the headers from the predicted
            sequences back to their parent sequence

    Yields:
        annotation: annotation for each *blast_gff* back translated

    """
    if split_func is None:
        def split_func(x): return tuple(x.rsplit('_', 1))

    prodigal_gff = {}
    for annotation in parse_gff(main_gff, strict=False):
        key = (
            annotation.seq_id,
            split_func(annotation.get_attr(attr))[1]
        )
        prodigal_gff[key] = (
            annotation.start,
            annotation.end,
            annotation.strand,
            annotation.get_attr(attr)
        )
    for annotation in parse_gff(blast_gff):
        key = split_func(annotation.seq_id)
        annotation.set_attr('prodigal_start', annotation.start)
        annotation.set_attr('prodigal_end', annotation.end)
        annotation.set_attr('prodigal_strand', annotation.strand)

        start, end, strand, p_id = prodigal_gff[key]

        annotation.seq_id = key[0]
        annotation.start = start
        annotation.end = end
        annotation.strand = strand
        annotation.set_attr('prodigal_ID', p_id)
        yield annotation



import pytest
import mgkit
import sys


def test_cite1(capsys):
    mgkit.cite(file_handle=sys.stderr)
    captured = capsys.readouterr()
    result = '\n'.join(
        [
            mgkit.LOGO,
            'MGKit Version: {0}'.format(mgkit.__VERSION__),
            mgkit.CITE
        ]
    )
    assert captured.err == result


def test_cite2(capsys):
    mgkit.cite(file_handle=sys.stdout)
    captured = capsys.readouterr()
    result = '\n'.join(
        [
            mgkit.LOGO,
            'MGKit Version: {0}'.format(mgkit.__VERSION__),
            mgkit.CITE
        ]
    )
    assert captured.out == result


def test_check_version():
    assert mgkit.check_version(mgkit.__VERSION__) is None


def test_dependency_error():
    with pytest.raises(mgkit.DependencyError):
        raise mgkit.DependencyError('test')
